# Name:  exporter.py
# Purpose: Container of Exporter class
# Authors:      Anton Korosov, Artem Moiseev
# Created:      22.01.2018
# Copyright:    (c) NERSC 2011 - 2018
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import print_function, absolute_import, division

import os
import tempfile
import datetime
import warnings

import gdal
import numpy as np

from netCDF4 import Dataset

from nansat.vrt import VRT
from nansat.node import Node

from nansat.warnings import NansatFutureWarning
from nansat.exceptions import NansatGDALError


class Exporter(object):
    """Abstract class for export functions """
    EXPORT_FILENAME_WARNING = ('Nansat.export(fileName=...) will be disabled from Nansat 1.1. '
                               'Use Nansat.export(filename=...).')
    EXPORT_RM_METADATA_WARNING = ('Nansat.export(rmMetadata=...) will be disabled from Nansat 1.1. '
                                  'Use Nansat.export(rm_metadata=...).')
    EXPORT_ADD_GEOLOC_WARNING = ('Nansat.export(addGeoloc=...) will be disabled from Nansat 1.1. '
                                 'Use Nansat.export(add_geolocation=...).')
    EXPORT_ADD_GCPS_WARNING = ('Nansat.export(addGCPs=...) has no effect and '
                               'will be disabled from Nansat 1.1. ')
    EXPORT_BOTTOMUP_WARNING = ('Nansat.export(bottomup=...) will be disabled from Nansat 1.1. '
                               'Use Nansat.export(options=[WRITE_BOTTOMUP=NO]')
    EXPORT_CREATED_WARNING = ('Nansat.export2thredds(createdTime=...) will be disabled from Nansat 1.1. '
                               'Use Nansat.export2thredds(created=...')
    EXPORT_MASKNAME_WARNING = ('Nansat.export2thredds(maskName=...) will be disabled from Nansat 1.1. '
                               'Use Nansat.export2thredds(mask_name=...')
    DEFAULT_INSTITUTE = 'NERSC'
    DEFAULT_SOURCE = 'satellite remote sensing'

    UNWANTED_METADATA = ['dataType', 'SourceFilename', 'SourceBand', '_Unsigned', 'FillValue',
                                'time', '_FillValue', 'type', 'scale', 'offset']

    def export(self, filename='', fileName='', bands=None, rm_metadata=None, rmMetadata=None,
               addGeoloc=None, add_geolocation=True,
               addGCPs=None, driver='netCDF', bottomup=None, options=None, hardcopy=False):
        '''Export Nansat object into netCDF or GTiff file

        Parameters
        -----------
        filename : str
            output file name
        bands: list (default=None)
            Specify band numbers to export.
            If None, all bands are exported.
        rm_metadata : list
            metadata names for removal before export.
            e.g. ['name', 'colormap', 'source', 'sourceBands']
        add_geolocation : bool
            add geolocation array datasets to exported file?
        driver : str
            Name of GDAL driver (format)
        bottomup : bool
            False: Write swath-projected data with rows and columns
                   organized as in the original product.
            True:  Use the default behaviour of GDAL, which is to flip the rows
        options : str or list
            GDAL export options in format of: 'OPT=VAL', or
            ['OPT1=VAL1', 'OP2='VAL2']
            See also http://www.gdal.org/frmt_netcdf.html
        hardcopy : bool
            Evaluate all bands just before export?

        Modifies
        ---------
        Create a netCDF file

        Notes
        ------
        If number of bands is more than one, serial numbers are added at the end of each band name.
        It is possible to fix it by changing line.4605 in GDAL/frmts/netcdf/netcdfdataset.cpp :
        'if( nBands > 1 ) sprintf(szBandName,"%s%d",tmpMetadata,iBand);'
        --> 'if( nBands > 1 ) sprintf(szBandName,"%s",tmpMetadata);'

        CreateCopy fails in case the band name has special characters,
        e.g. the slash in 'HH/VV'.

        Metadata strings with special characters are escaped with XML/HTML
        encoding.

        Examples
        --------
        # export all the bands into a netDCF 3 file
        >>> n.export(netcdfile)

        # export all bands into a GeoTiff
        >>> n.export(driver='GTiff')

        '''
        # trigger Nansat future warnings on deprecated features
        if fileName != '':
            warnings.warn(self.EXPORT_FILENAME_WARNING, NansatFutureWarning)
            filename = fileName
        if rmMetadata is not None:
            warnings.warn(self.EXPORT_RM_METADATA_WARNING, NansatFutureWarning)
            rm_metadata = rmMetadata
        if addGeoloc is not None:
            warnings.warn(self.EXPORT_ADD_GEOLOC_WARNING, NansatFutureWarning)
            add_geolocation = addGeoloc
        if addGCPs is not None:
            warnings.warn(self.EXPORT_ADD_GCPS_WARNING, NansatFutureWarning)
        if isinstance(bottomup, bool):
            warnings.warn(self.EXPORT_BOTTOMUP_WARNING, NansatFutureWarning)

        if options is None:
            options = []
        if type(options) == str:
            options = [options]

        # temporary VRT for exporting
        export_vrt = self.vrt.copy()
        export_vrt.leave_few_bands(bands)
        export_vrt.split_complex_bands()
        if add_geolocation:
            export_vrt.create_geolocation_bands()
        export_vrt.fix_band_metadata(rm_metadata)
        export_vrt.fix_global_metadata(rm_metadata)

        # if output filename is the same as input one
        if self.filename == filename or hardcopy:
            export_vrt.hardcopy_bands()

        if driver == 'GTiff':
            options, add_gcps = export_vrt.prepare_export_gtiff(options)
        else:
            options, add_gcps = export_vrt.prepare_export_netcdf(options, bottomup)

        # Create output file using GDAL
        dataset = gdal.GetDriverByName(driver).CreateCopy(filename, export_vrt.dataset, options=options)
        del dataset
        # add GCPs into netCDF file as separate float variables
        if add_gcps:
            Exporter._add_gcps(filename, export_vrt.dataset.GetGCPs(), bottomup)

        self.logger.debug('Export - OK!')

    def export2thredds(self, filename, bands, metadata=None,
                        mask_name=None, maskName=None, rm_metadata=None, rmMetadata=None,
                        time=None, created=None, createdTime=None):
        ''' Export data into a netCDF formatted for THREDDS server

        Parameters
        ----------
        filename : str
            output file name
        bands : dict
            {'band_name': {'type'     : '>i1',
                           'scale'    : 0.1,
                           'offset'   : 1000,
                           'metaKey1' : 'meta value 1',
                           'metaKey2' : 'meta value 2'}}
            dictionary sets parameters for band creation
            'type' - string representation of data type in the output band
            'scale' - sets scale_factor and applies scaling
            'offset' - sets 'scale_offset and applies offsetting
            other entries (e.g. 'units': 'K') set other metadata
        metadata : dict
            Glbal metadata to add
        mask_name: str
            if data include a mask band: give the mask name.
            Non-masked value is 64.
            if None: no mask is added
        rm_metadata : list
            unwanted metadata names which will be removed
        time : list with datetime objects
            aqcuisition time of original data. That value will be in time dim
        created : datetime
            date of creation. Will be in metadata 'created'

        Note
        ----
        Nansat object (self) has to be projected (with valid GeoTransform and
        valid Spatial reference information) but not wth GCPs

        Examples
        --------
        # create THREDDS formatted netcdf file with all bands and time variable
        >>> n.export2thredds(filename)

        # export only the first band and add global metadata
        >>> n.export2thredds(filename, ['L_469'], {'description': 'example'})

        # export several bands and modify type, scale and offset
        >>> bands = {'L_645' : {'type': '>i2', 'scale': 0.1, 'offset': 0},
                     'L_555' : {'type': '>i2', 'scale': 0.1, 'offset': 0}}
        >>> n.export2thredds(filename, bands)

        '''

        if rmMetadata is not None:
            warnings.warn(self.EXPORT_RM_METADATA_WARNING, NansatFutureWarning)
            rm_metadata = rmMetadata
        if createdTime is not None:
            warnings.warn(self.EXPORT_CREATED_WARNING, NansatFutureWarning)
            created = createdTime
        if maskName is not None:
            warnings.warn(self.EXPORT_MASKNAME_WARNING, NansatFutureWarning)
            mask_name = maskName

        if not isinstance(bands, dict):
            raise ValueError('<bands> must be dict!')
        if metadata is None:
            metadata = {}
        # raise error if self is not projected (has GCPs)
        if len(self.vrt.dataset.GetGCPs()) > 0:
            raise ValueError('Cannot export dataset with GCPS for THREDDS!')

        # Create temporary empty Nansat object with self domain
        data = self.__class__.__new__(self.__class__)
        data._init_from_domain(self)

        # get mask (if exist)
        if mask_name is not None:
            mask = self[mask_name]

        # add required bands to data
        dstBands = {}
        srcBands = [self.bands()[b]['name'] for b in self.bands()]
        for iband in bands:
            # skip non exiting bands
            if iband not in srcBands:
                self.logger.error('%s is not found' % str(iband))
                continue

            array = self[iband]

            # catch None band error
            if array is None:
                raise NansatGDALError('%s is None' % str(iband))

            # set type, scale and offset from input data or by default
            dstBands[iband] = {}
            dstBands[iband]['type'] = bands[iband].get('type',
                                             array.dtype.str.replace('u', 'i'))
            dstBands[iband]['scale'] = float(bands[iband].get('scale', 1.0))
            dstBands[iband]['offset'] = float(bands[iband].get('offset', 0.0))
            if '_FillValue' in bands[iband]:
                dstBands[iband]['_FillValue'] = np.array(
                                            [bands[iband]['_FillValue']],
                                            dtype=dstBands[iband]['type'])[0]

            # mask values with np.nan
            if maskName is not None and iband != maskName:
                array[mask != 64] = np.nan

            # add array to a temporary Nansat object
            bandMetadata = self.get_metadata(band_id=iband)
            data.add_band(array=array, parameters=bandMetadata)
        self.logger.debug('Bands for export: %s' % str(dstBands))

        global_metadata = Exporter._set_global_metadata(created, data, metadata)

        # export temporary Nansat object to a temporary netCDF
        fid, tmp_filename = tempfile.mkstemp(suffix='.nc')
        os.close(fid)
        data.export(tmp_filename, rm_metadata=rm_metadata)
        del data

        self._post_proc_thredds(tmp_filename, filename, bands, dstBands, time, global_metadata)

    def _create_dimensions(self, nc_inp, nc_out, time):
        """Create space and time dimenstions in the destination file"""
        # get time from Nansat object or from input datetime
        if time is None:
            time = self.time_coverage_start

        # collect info on dimension names
        dim_names = []
        grid_mapping_name = None
        for inp_var_name in nc_inp.variables:
            inp_var = nc_inp.variables[inp_var_name]
            dim_names += list(inp_var.dimensions)
            # get grid_mapping_name
            if hasattr(inp_var, 'grid_mapping_name'):
                grid_mapping_name = inp_var.grid_mapping_name
                grid_mapping_var_name = inp_var_name
        dim_names = list(set(dim_names))

        # collect info on dimension shapes
        dim_shapes = {}
        for dim_name in dim_names:
            dim_var = nc_inp.variables[dim_name]
            dim_shapes[dim_name] = dim_var.shape[0]

        # create dimensions
        for dim_name in dim_names:
            nc_out.createDimension(dim_name, dim_shapes[dim_name])

        # add time dimension
        nc_out.createDimension('time', 1)
        out_var = nc_out.createVariable('time', '>f8',  ('time', ))
        out_var.calendar = 'standard'
        out_var.long_name = 'time'
        out_var.standard_name = 'time'
        out_var.units = 'days since 1900-1-1 0:0:0 +0'
        out_var.axis = 'T'
        # create value for time variable
        td = time - datetime.datetime(1900, 1, 1)
        days = td.days + (float(td.seconds) / 60.0 / 60.0 / 24.0)
        # add date
        out_var[:] = days

        return grid_mapping_name, grid_mapping_var_name

    def _post_proc_thredds(self, tmp_filename, out_filename, bands, band_metadata, time, global_metadata):
        """Post processing of file for THREDDS (add time variable and metadata)"""
        # open files for input and output
        nc_inp = Dataset(tmp_filename, 'r')
        nc_out = Dataset(out_filename, 'w')

        grid_mapping_name, grid_mapping_var_name = self._create_dimensions(nc_inp, nc_out, time)

        # recreate file
        for inp_var_key in nc_inp.variables.keys():
            inp_var = nc_inp.variables[inp_var_key]
            if 'name' in inp_var.ncattrs():
                inp_var_name = inp_var.getncattr('name')
            else:
                inp_var_name = inp_var.name
            # create projection var
            if inp_var_name == grid_mapping_var_name:
                out_var = Exporter._copy_nc_var(inp_var, nc_out, grid_mapping_name,
                                                inp_var.dtype.str, inp_var.dimensions)
                continue

            # create simple x/y variables
            if inp_var_name in ['x', 'y', 'lon', 'lat']:
                out_var = Exporter._copy_nc_var(inp_var, nc_out, inp_var_name,
                                                '>f4', inp_var.dimensions)
            # create data var
            elif inp_var_name in band_metadata:
                fill_value = None
                if '_FillValue' in inp_var.ncattrs():
                    fill_value = inp_var._FillValue
                if '_FillValue' in band_metadata[inp_var_name]:
                    fill_value = band_metadata['_FillValue']
                dimensions = ('time', ) + inp_var.dimensions
                out_var = Exporter._copy_nc_var(inp_var, nc_out, inp_var_name,
                                                band_metadata[inp_var_name]['type'],
                                                dimensions, fill_value=fill_value)

            # copy array from input data
            data = inp_var[:]

            # copy rounded data from x/y
            if inp_var_name in ['x', 'y']:
                out_var[:] = np.floor(data).astype('>f4')
                # add axis=X or axis=Y
                out_var.axis = {'x': 'X', 'y': 'Y'}[inp_var_name]

            # copy data from lon/lat
            if inp_var_name in ['lon', 'lat']:
                out_var[:] = data.astype('>f4')

            # copy data from variables in the list
            if inp_var_name in band_metadata:
                # add offset and scale attributes
                scale = band_metadata[inp_var_name]['scale']
                offset = band_metadata[inp_var_name]['offset']
                if not (offset == 0.0 and scale == 1.0):
                    out_var.setncattr('add_offset', offset)
                    out_var.setncattr('scale_factor', scale)
                    data = (data - offset) / scale

                out_var[:] = data.astype(band_metadata[inp_var_name]['type'])

                # add custom attributes from input parameter bands
                if inp_var_name in bands:
                    for newAttr in bands[inp_var_name]:
                        if newAttr not in Exporter.UNWANTED_METADATA:
                            out_var.setncattr(newAttr, bands[inp_var_name][newAttr])
                    # add grid_mapping info
                    if grid_mapping_name is not None:
                        out_var.setncattr('grid_mapping', grid_mapping_name)

        # copy (some) global attributes
        for globAttr in nc_inp.ncattrs():
            if not(globAttr.strip().startswith('GDAL')):
                nc_out.setncattr(globAttr, nc_inp.getncattr(globAttr))

        # add common and custom global attributes
        nc_out.setncatts(global_metadata)

        # write output file
        nc_out.close()

        # close original files
        nc_inp.close()

        # Delete the temprary netCDF file
        os.remove(tmp_filename)

        return 0

    @staticmethod
    def _set_global_metadata(created, data, metadata):
        if created is None:
            created = (datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC'))
        # get corners of reprojected data
        min_lon, max_lon, min_lat, max_lat = data.get_min_max_lon_lat()

        global_metadata = {
                    'institution': Exporter.DEFAULT_INSTITUTE,
                    'source': Exporter.DEFAULT_SOURCE,
                    'creation_date': created,
                    'northernmost_latitude': np.float(max_lat),
                    'southernmost_latitude': np.float(min_lat),
                    'westernmost_longitude': np.float(min_lon),
                    'easternmost_longitude': np.float(max_lon),
                    'history': ' '}
        global_metadata.update(metadata)

        return global_metadata

    @staticmethod
    def _add_gcps(filename, gcps, bottomup):
        """Add 4 variables with gcps to the generated netCDF file"""
        gcp_variables = ['GCPX', 'GCPY', 'GCPZ', 'GCPPixel', 'GCPLine', ]

        # check if file exists
        if not os.path.exists(filename):
            warnings.warn('Cannot add GCPs! File %s doesn''t exist!' % filename)

        # open output file for adding GCPs
        ncFile = Dataset(filename, 'a')

        # get GCP values into single array from GCPs
        gcp_values = np.zeros((5, len(gcps)))
        for i, gcp in enumerate(gcps):
            gcp_values[0, i] = gcp.GCPX
            gcp_values[1, i] = gcp.GCPY
            gcp_values[2, i] = gcp.GCPZ
            gcp_values[3, i] = gcp.GCPPixel
            gcp_values[4, i] = gcp.GCPLine

        # make gcps dimensions
        ncFile.createDimension('gcps', len(gcps))
        # make gcps variables and add data
        for i, var in enumerate(gcp_variables):
            var = ncFile.createVariable(var, 'f4', ('gcps',))
            var[:] = gcp_values[i]

        # write data, close file
        ncFile.close()

    @staticmethod
    def _copy_nc_var(inp_var, nc_out, var_name, var_type, dimensions, fill_value=None):
        """Create new NC variable, set name, type, dimensions and copy attributes"""
        out_var = nc_out.createVariable(var_name, var_type, dimensions, fill_value=fill_value)
        for ncattr in inp_var.ncattrs():
            if str(ncattr) not in Exporter.UNWANTED_METADATA:
                out_var.setncattr(str(ncattr), inp_var.getncattr(ncattr))

        return out_var



# ------------------------------------------------------------------------------
# Name:         test_nansat.py
# Purpose:      Test the Nansat class
#
# Author:       Morten Wergeland Hansen, Asuka Yamakawa, Anton Korosov
#
# Created:      18.06.2014
# Last modified:24.08.2017 14:00
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
# ------------------------------------------------------------------------------
from __future__ import print_function, absolute_import, division
import os
import sys
import json
import unittest
import warnings
import datetime
from xml.sax.saxutils import unescape
from mock import patch, PropertyMock, Mock, MagicMock, DEFAULT

import numpy as np

try:
    if 'DISPLAY' not in os.environ:
        import matplotlib; matplotlib.use('Agg')
    import matplotlib
    import matplotlib.pyplot as plt
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True

import gdal
from netCDF4 import Dataset

from nansat import Nansat, Domain, NSR
from nansat.warnings import NansatFutureWarning
from nansat.tests.nansat_test_base import NansatTestBase

warnings.simplefilter("always", NansatFutureWarning)
warnings.simplefilter("always", UserWarning)


class ExporterTest(NansatTestBase):

    def test_geolocation_of_exportedNC_vs_original(self):
        """ Lon/lat in original and exported file should coincide """
        orig = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        orig.export(self.tmp_filename)

        copy = Nansat(self.tmp_filename, mapper=self.default_mapper)
        lon0, lat0 = orig.get_geolocation_grids()
        lon1, lat1 = copy.get_geolocation_grids()
        np.testing.assert_allclose(lon0, lon1)
        np.testing.assert_allclose(lat0, lat1)

    def test_special_characters_in_exported_metadata(self):
        orig = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        orig.vrt.dataset.SetMetadataItem('jsonstring', json.dumps({'meta1':
                                         'hei', 'meta2': 'derr'}))
        orig.export(self.tmp_filename)
        copy = Nansat(self.tmp_filename, mapper=self.default_mapper)
        dd = json.loads(unescape(copy.get_metadata('jsonstring'), {'&quot;':
                                                                   '"'}))
        self.assertIsInstance(dd, dict)

    def test_time_coverage_metadata_of_exported_equals_original(self):
        orig = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        orig.set_metadata('time_coverage_start', '2010-01-02T08:49:02.347809')
        orig.set_metadata('time_coverage_end', '2010-01-02T08:50:03.599373')
        orig.export(self.tmp_filename)
        copy = Nansat(self.tmp_filename, mapper=self.default_mapper)

        self.assertEqual(orig.get_metadata('time_coverage_start'),
                copy.get_metadata('time_coverage_start'))
        self.assertEqual(orig.get_metadata('time_coverage_end'),
                copy.get_metadata('time_coverage_end'))

    def test_export_netcdf(self):
        """ Test export and following import of data with bands containing np.nan values """
        n = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        arrNoNaN = np.random.randn(n.shape()[0], n.shape()[1])
        n.add_band(arrNoNaN, {'name': 'testBandNoNaN'})
        arrWithNaN = arrNoNaN.copy()
        arrWithNaN[int(n.shape()[0] / 2.) - 10:int(n.shape()[0] / 2 + 10),
                   int(n.shape()[1] / 2.) - 10:int(n.shape()[1] / 2 + 10)] = np.nan
        n.add_band(arrWithNaN, {'name': 'testBandWithNaN'})
        n.export(self.tmp_filename)
        exported = Nansat(self.tmp_filename, mapper=self.default_mapper)
        earrNoNaN = exported['testBandNoNaN']
        # Use allclose to allow some roundoff errors
        self.assertTrue(np.allclose(arrNoNaN, earrNoNaN))
        earrWithNaN = exported['testBandWithNaN']
        np.testing.assert_allclose(arrWithNaN, earrWithNaN)

    def test_export_gcps_filename_warning(self):
        """ Should export file with GCPs and write correct bands"""
        n0 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'temp.nc')
        with warnings.catch_warnings(record=True) as w:
            n0.export(fileName=tmpfilename)
            self.assertEqual(len(w), 1)
            self.assertIn('Nansat.export(fileName', str(w[0].message))

    def test_export_gcps_to_netcdf(self):
        """ Should export file with GCPs and write correct bands"""
        n0 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_export_gcps.nc')
        n0.export(tmpfilename)

        ncf = Dataset(tmpfilename)
        self.assertTrue(os.path.exists(tmpfilename))
        self.assertTrue('GCPX' in ncf.variables)
        self.assertTrue('GCPY' in ncf.variables)
        self.assertTrue('GCPPixel' in ncf.variables)
        self.assertTrue('GCPLine' in ncf.variables)

        n1 = Nansat(tmpfilename, mapper=self.default_mapper)
        b0 = n0['L_469']
        b1 = n1['L_469']
        np.testing.assert_allclose(b0, b1)

        lon0, lat0 = n0.get_geolocation_grids()
        lon1, lat1 = n1.get_geolocation_grids()
        np.testing.assert_allclose(lon0, lon1)
        np.testing.assert_allclose(lat0, lat1)

    def test_export_gcps_complex_to_netcdf(self):
        """ Should export file with GCPs and write correct complex bands"""
        n0 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        b0 = n0['L_469']

        n1 = Nansat.from_domain(n0)
        n1.add_band(b0.astype('complex64'), parameters={'name': 'L_469'})

        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_export_gcps_complex.nc')
        n1.export(tmpfilename)

        ncf = Dataset(tmpfilename)
        self.assertTrue(os.path.exists(tmpfilename))
        self.assertTrue('GCPX' in ncf.variables)
        self.assertTrue('GCPY' in ncf.variables)
        self.assertTrue('GCPPixel' in ncf.variables)
        self.assertTrue('GCPLine' in ncf.variables)

        n2 = Nansat(tmpfilename, mapper=self.default_mapper)
        b2 = n2['L_469']

        lon0, lat0 = n0.get_geolocation_grids()
        lon2, lat2 = n1.get_geolocation_grids()
        np.testing.assert_allclose(lon0, lon2)
        np.testing.assert_allclose(lat0, lat2)

    def test_export_gtiff(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_export.tif')
        n.export(tmpfilename, driver='GTiff')

        self.assertTrue(os.path.exists(tmpfilename))

    def test_export_band(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export_band.tif')
        n.export(tmpfilename, bands=[1], driver='GTiff')
        n = Nansat(tmpfilename, mapper=self.default_mapper)

        self.assertTrue(os.path.exists(tmpfilename))
        self.assertEqual(n.vrt.dataset.RasterCount, 1)

    def test_export_band_by_name(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export_band.tif')
        n.export(tmpfilename, bands=['L_645'], driver='GTiff')
        n = Nansat(tmpfilename, mapper=self.default_mapper)

        self.assertTrue(os.path.exists(tmpfilename))
        self.assertEqual(n.vrt.dataset.RasterCount, 1)

    def test_reproject_and_export_band(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n2 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n1.reproject(n2)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_reproject_export_band.nc')
        n1.export(tmpfilename, bands=[1])

        n = Nansat(tmpfilename, mapper=self.default_mapper)
        self.assertTrue(os.path.exists(tmpfilename))
        self.assertEqual(n.vrt.dataset.RasterCount, 1)

    def test_export_selected_bands(self):
        n = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        resfile = 'tmp.nc'
        new_band = np.random.randn(n.shape()[0], n.shape()[1])
        n.add_band(new_band, {'name': 'newBand'})
        # Test with band numbers
        n.export(resfile, bands=[4, 2])
        self.assertTrue(os.path.exists(resfile))
        nn = Nansat(resfile, mapper=self.default_mapper)
        self.assertTrue(nn.has_band('newBand'))
        self.assertTrue(nn.has_band('L_555'))
        os.unlink(resfile)
        # Test with band names - not yet implemented
#         n.export(resfile, bands=['newBand', 'L_555'])
#         nn = Nansat(resfile)
#         self.assertTrue(nn.has_band('newBand'))
#         self.assertTrue(nn.has_band('L_555'))
#         os.unlink(resfile)

    def test_export_option(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export_option.nc')
        # Test with band numbers
        n.export(tmpfilename, options='WRITE_LONLAT=YES')
        n.export(tmpfilename + '2', options=['WRITE_LONLAT=NO'])
        nn = Nansat(tmpfilename, mapper=self.default_mapper)
        nn2 = Nansat(tmpfilename + '2', mapper=self.default_mapper)
        self.assertTrue(nn.has_band('lon'))
        self.assertTrue(nn.has_band('lat'))
        self.assertTrue(nn.has_band('Bristol'))
        self.assertFalse(nn2.has_band('lon'))
        self.assertFalse(nn2.has_band('lat'))
        self.assertTrue(nn2.has_band('Bristol'))

    def test_export2thredds_arctic_long_lat(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper, log_level=40)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export2thredds_arctic.nc')
        bands = {
            'Bristol': {'type': '>i2'},
            'Bootstrap': {'type': '>i2'},
            'UMass_AES': {'type': '>i2'},
        }
        n.export2thredds(tmpfilename, bands, time=datetime.datetime(2016, 1, 20))

        self.assertTrue(os.path.exists(tmpfilename))
        g = gdal.Open(tmpfilename)
        metadata = g.GetMetadata_Dict()

		# GDAL behaves differently:
		# Windows: nc-attributes are accessible without 'NC_GLOBAL#' prefix
		# Linux: nc-attributes are accessible only with 'NC_GLOBAL#' prefix
        # OSX: ?
        # Therefore we have to add NC_GLOBAL# and test if such metadata exists
        nc_prefix = 'NC_GLOBAL#'
        if not nc_prefix + 'easternmost_longitude' in metadata:
            nc_prefix = ''
        self.assertIn(nc_prefix + 'easternmost_longitude', metadata)

        # Test that the long/lat values are set correctly
        test_metadata_keys = ['easternmost_longitude', 'westernmost_longitude',
                              'northernmost_latitude', 'southernmost_latitude']
        test_metadata_min = [179, -180, 89.9, 53]
        test_metadata_max = [180, -179, 90, 54]
        for i, test_metadata_key in enumerate(test_metadata_keys):
            medata_value = float(metadata[nc_prefix + test_metadata_key])
            self.assertTrue(medata_value >= test_metadata_min[i],
                            '%s is wrong: %f'%(test_metadata_key, medata_value))
            self.assertTrue(medata_value <= test_metadata_max[i],
                            '%s is wrong: %f'%(test_metadata_key, medata_value))

    def test_dont_export2thredds_gcps(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n2 = Nansat.from_domain(n)
        n.add_band(np.ones(n2.shape(), np.float32))
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export2thredds.nc')
        self.assertRaises(ValueError, n2.export2thredds, tmpfilename,
                          ['L_645'])

    def test_export2thredds_longlat_list(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        with self.assertRaises(ValueError):
            n.export2thredds('aa', ['L_469'])

    def test_export2thredds_longlat_dict(self):
        d = Domain("+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs",
                   "-te 27 70 31 72 -ts 200 200")
        n = Nansat.from_domain(d)
        n.add_band(np.ones(d.shape(), np.float32),
                   parameters={'name': 'L_469'})
        n.set_metadata('time_coverage_start', '2016-01-19')

        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export2thredds_longlat.nc')
        n.export2thredds(tmpfilename, {'L_469': {'type': '>i1'}})
        ncI = Dataset(tmpfilename, 'r')
        ncIVar = ncI.variables['L_469']
        self.assertTrue(ncIVar.grid_mapping in ncI.variables.keys())
        self.assertEqual(ncIVar[:].dtype, np.int8)


    def test_export_netcdf_complex_remove_meta(self):
        n = Nansat(self.test_file_complex, mapper=self.default_mapper)
        self.assertEqual(n.get_metadata('PRODUCT_TYPE'), 'SLC')
        with warnings.catch_warnings(record=True) as recorded_warnings:
            n.export(self.tmp_filename, rmMetadata=['PRODUCT_TYPE'])
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)
        exported = Nansat(self.tmp_filename, mapper=self.default_mapper)
        with self.assertRaises(ValueError):
            exported.get_metadata('PRODUCT_TYPE')
        self.assertTrue((n[1] == exported[1]).any())

    def test_export_netcdf_arctic(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        n.export(self.tmp_filename)
        exported = Nansat(self.tmp_filename, mapper=self.default_mapper)
        self.assertTrue((n[1] == exported[1]).any())
        self.assertTrue((n[2] == exported[2]).any())
        self.assertTrue((n[3] == exported[3]).any())

    def test_export_netcdf_arctic_hardcopy(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        n.export(self.tmp_filename, hardcopy=True)
        exported = Nansat(self.tmp_filename, mapper=self.default_mapper)
        self.assertTrue((n[1] == exported[1]).any())
        self.assertTrue((n[2] == exported[2]).any())
        self.assertTrue((n[3] == exported[3]).any())

    @patch('nansat.exporter.VRT._add_geolocation')
    def test_export_add_geoloc(self, mock_add_geolocation):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        with warnings.catch_warnings(record=True) as recorded_warnings:
            n.export(self.tmp_filename, addGeoloc=True)
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)
        self.assertTrue(mock_add_geolocation.called)

    def test_export_add_gcps(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        with warnings.catch_warnings(record=True) as recorded_warnings:
            n.export(self.tmp_filename, addGCPs=True, bottomup=True)
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)
            self.assertEqual(recorded_warnings[1].category, NansatFutureWarning)

    def test_export2thredds_rmmetadata(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper, log_level=40)
        with warnings.catch_warnings(record=True) as recorded_warnings:
            n.export2thredds(self.tmp_filename, {'Bristol': {'type': '>i2'}},
                            time=datetime.datetime(2016, 1, 20),
                            rmMetadata=['description'])
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)

if __name__ == "__main__":
    unittest.main()



# Name:        nansat_mapper_merisL2
# Purpose:     Mapping for Meris-L2 data
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from dateutil.parser import parse

from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError
from envisat import Envisat


class Mapper(VRT, Envisat):
    ''' Create VRT with mapping of WKV for MERIS Level 2 (FR or RR)'''

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 geolocation=False, zoomSize=500, step=1, **kwargs):

        ''' Create MER2 VRT

        Parameters
        -----------
        filename : string
        gdalDataset : gdal dataset
        gdalMetadata : gdal metadata
        geolocation : bool (default is False)
            if True, add gdal geolocation
        zoomSize: int (used in envisat.py)
            size, to which the ADS array will be zoomed using scipy
            array of this size will be stored in memory
        step: int (used in envisat.py)
            step of pixel and line in GeolocationArrays. lat/lon grids are
            generated at that step
        '''

        self.setup_ads_parameters(filename, gdalMetadata)

        if self.product[0:9] != "MER_FRS_2" and self.product[0:9] != "MER_RR__2":
            raise WrongMapperError(filename)

        metaDict = [{'src': {'SourceFilename': filename, 'SourceBand': 1},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '412'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 2},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '443'}},
                    {'src': {'SourceFilename': filename, 'SourceBand':  3},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '490'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 4},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '510'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 5},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '560'}},
                    {'src': {'SourceFilename': filename, 'SourceBand':  6},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '620'}},
                    {'src': {'SourceFilename': filename, 'SourceBand':  7},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '665'}},
                    {'src': {'SourceFilename': filename, 'SourceBand':  8},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '680'}},
                    {'src': {'SourceFilename': filename, 'SourceBand':  9},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '708'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 10},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '753'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 11},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '761'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 12},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '778'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 13},
                     'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air',
                             'wavelength': '864'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 15},
                     'dst': {'wkv': 'mass_concentration_of_chlorophyll_a_in_sea_water',
                             'suffix': '1_log', 'case': 'I'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 16},
                     'dst': {'wkv': 'volume_absorption_coefficient_of_radiative_flux_in_sea_water_due_to_dissolved_organic_matter',
                             'suffix': '2_log', 'case': 'II'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 17},
                     'dst': {'wkv': 'mass_concentration_of_suspended_matter_in_sea_water',
                             'suffix': '2_log', 'case': 'II'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 18},
                     'dst': {'wkv': 'mass_concentration_of_chlorophyll_a_in_sea_water',
                             'suffix': '2_log', 'case': 'II'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 22},
                     'dst': {'wkv': 'quality_flags', 'suffix': 'l2'}}
                    ]

        # add 'name' to 'parameters'
        for bandDict in metaDict:
            if 'wavelength' in bandDict['dst']:
                bandDict['dst']['suffix'] = bandDict['dst']['wavelength']

        #get GADS from header
        scales = self.read_scaling_gads(range(7, 20) + [20, 21, 22, 20])
        offsets = self.read_scaling_gads(range(33, 46) + [46, 47, 48, 46])
        # set scale/offset to the band metadata (only reflectance)
        for i, bandDict in enumerate(metaDict[:-1]):
            bandDict['src']['ScaleRatio'] = str(scales[i])
            bandDict['src']['ScaleOffset'] = str(offsets[i])

        # add log10-scaled variables
        metaDict += [{'src': {'SourceFilename': filename, 'SourceBand': 1},
                      'dst': {'wkv': 'mass_concentration_of_chlorophyll_a_in_sea_water',
                              'suffix': '1', 'case': 'I',
                              'expression': 'np.power(10., self["chlor_a_1_log"])'}},
                     {'src': {'SourceFilename': filename, 'SourceBand': 1},
                      'dst': {'wkv': 'mass_concentration_of_chlorophyll_a_in_sea_water',
                              'suffix': '2', 'case': 'II',
                              'expression': 'np.power(10., self["chlor_a_2_log"])'}},
                     {'src': {'SourceFilename': filename, 'SourceBand': 1},
                      'dst': {'wkv': 'volume_absorption_coefficient_of_radiative_flux_in_sea_water_due_to_dissolved_organic_matter',
                              'suffix': '2', 'case': 'II',
                              'expression': 'np.power(10., self["cdom_a_2_log"])'}},
                     {'src': {'SourceFilename': filename, 'SourceBand': 1},
                      'dst': {'wkv': 'mass_concentration_of_suspended_matter_in_sea_water',
                              'suffix': '2', 'case': 'II',
                              'expression': 'np.power(10., self["tsm_2_log"])'}}
                     ]

        # get list with resized VRTs from ADS
        self.band_vrts = {'adsVRTs': self.get_ads_vrts(gdalDataset,
                                                     ['sun zenith angles',
                                                      'sun azimuth angles',
                                                      'zonal winds',
                                                      'meridional winds'],
                                                     zoomSize=zoomSize,
                                                     step=step)}

        # add bands from the ADS VRTs
        for adsVRT in self.band_vrts['adsVRTs']:
            metaDict.append({'src': {'SourceFilename': adsVRT.filename,
                                     'SourceBand': 1},
                             'dst': {'name': (adsVRT.dataset.GetRasterBand(1).
                                              GetMetadataItem('name')),
                                     'units': (adsVRT.dataset.GetRasterBand(1).
                                               GetMetadataItem('units'))}
                             })

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalDataset)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # set time
        self._set_envisat_time(gdalMetadata)

        # add geolocation arrays
        if geolocation:
            self.add_geolocation_from_ads(gdalDataset,
                                          zoomSize=zoomSize, step=step)
        # set time
        self._set_envisat_time(gdalMetadata)

        self.dataset.SetMetadataItem('sensor', 'MERIS')
        self.dataset.SetMetadataItem('satellite', 'ENVISAT')



#-------------------------------------------------------------------------------
# Name:		mapper_amsre_UHAM_lead_fraction.py
# Purpose:
#
# Author:       Morten Wergeland Hansen
# Modified:	Morten Wergeland Hansen
#
# Created:	18.02.2015
# Last modified:24.02.2015 09:26
# Copyright:    (c) NERSC
# License:
#-------------------------------------------------------------------------------
import datetime
from osgeo import gdal, osr
from nansat.nsr import NSR
from nansat.vrt import VRT

from nansat.exceptions import WrongMapperError

class Mapper(VRT):

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):

        title_correct = False
        if not gdalMetadata:
            raise WrongMapperError
        for key, val in list(gdalMetadata.items()):
            if 'title' in key:
                if not val == 'Daily AMSR-E Arctic lead area fraction [in percent]':
                    raise WrongMapperError
                else:
                    title_correct = True

        if not title_correct:
            raise WrongMapperError

        # initiate VRT for the NSIDC 10 km grid
        self._init_from_dataset_params(1216, 1792, (-3850000, 6250, 0.0, 5850000, 0.0, -6250),
                                       NSR(3411).wkt)

        src = {
            'SourceFilename': 'NETCDF:"%s":lf' % filename,
            'SourceBand': 1,
        }
        dst = {
            'name': 'leadFraction',
            'long_name': 'AMSRE sea ice lead fraction',
        }

        self.create_band(src, dst)
        self.dataset.FlushCache()





# Name:         mapper_ncep_wind
# Purpose:      Nansat mapping for NCEP GFS model data subset as downloaded
#               by mapper_ncep_wind_online
# Author:       Knut-Frode Dagestad
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
#
# Made for GRIB files downloaded from http://nomads.ncep.noaa.gov/data/gfs4/
import datetime
import json
import pythesint as pti

from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError


class Mapper(VRT):
    ''' VRT with mapping of WKV for NCEP GFS '''

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' Create NCEP VRT '''

        if not gdalDataset:
            raise WrongMapperError(filename)

        geotransform = gdalDataset.GetGeoTransform()
        if (geotransform != (-0.25, 0.5, 0.0, 90.25, 0.0, -0.5) or
                gdalDataset.RasterCount != 2):  # Not water proof
            raise WrongMapperError(filename)

        metaDict = [{'src': {'SourceFilename': filename,
                             'SourceBand': 1},
                     'dst': {'wkv': 'eastward_wind',
                             'height': '10 m'}},
                    {'src': {'SourceFilename': filename,
                             'SourceBand': 2},
                     'dst': {'wkv': 'northward_wind',
                             'height': '10 m'}},
                    {'src': [{'SourceFilename': filename,
                              'SourceBand': 1,
                              'DataType': gdalDataset.GetRasterBand(1).DataType
                              },
                             {'SourceFilename': filename,
                              'SourceBand': 2,
                              'DataType': gdalDataset.GetRasterBand(2).DataType
                              }],
                     'dst': {'wkv': 'wind_speed',
                             'PixelFunctionType': 'UVToMagnitude',
                             'name': 'windspeed',
                             'height': '2 m'
                             }},
                    {'src': [{'SourceFilename': filename,
                              'SourceBand': 1,
                              'DataType': gdalDataset.GetRasterBand(1).DataType
                              },
                             {'SourceFilename': filename,
                              'SourceBand': 2,
                              'DataType': gdalDataset.GetRasterBand(2).DataType
                              }],
                     'dst': {'wkv': 'wind_from_direction',
                             'PixelFunctionType': 'UVToDirectionFrom',
                             'name': 'winddirection',
                             'height': '2 m'
                             }
                     }]

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalDataset)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # Adding valid time from the GRIB file to dataset
        validTime = gdalDataset.GetRasterBand(1).GetMetadata()['GRIB_VALID_TIME']
        self.dataset.SetMetadataItem('time_coverage_start',
            (datetime.datetime.utcfromtimestamp(
                int(validTime.strip().split(' ')[0])).isoformat()))
        self.dataset.SetMetadataItem('time_coverage_end',
            (datetime.datetime.utcfromtimestamp(
                int(validTime.strip().split(' ')[0])).isoformat()))

        # Get dictionary describing the instrument and platform according to
        # the GCMD keywords
        mm = pti.get_gcmd_instrument('computer')
        ee = pti.get_gcmd_platform('ncep-gfs')

        # TODO: Validate that the found instrument and platform are indeed what we
        # want....

        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))



# Name:    tools.py
# Purpose: collection of data and funcs used in NANSAT modules
# Authors:      Asuka Yamakawa, Anton Korosov, Knut-Frode Dagestad,
#               Morten W. Hansen, Alexander Myasoyedov,
#               Dmitry Petrenko, Evgeny Morozov
# Created:      29.06.2011
# Copyright:    (c) NERSC 2011 - 2013
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import absolute_import

import os
import warnings
from nansat.warnings import NansatFutureWarning
import logging
from dateutil.parser import parse

try:
    if 'DISPLAY' not in os.environ:
        import matplotlib; matplotlib.use('Agg')
    from matplotlib import cm
    import matplotlib.pyplot as plt
    from matplotlib.colors import hex2color
    from mpl_toolkits.basemap import Basemap
    from matplotlib.patches import Polygon
except ImportError:
    BASEMAP_LIB_IS_INSTALLED = False
    MATPLOTLIB_IS_INSTALLED = False

else:
    BASEMAP_LIB_IS_INSTALLED = True
    MATPLOTLIB_IS_INSTALLED = True

import numpy as np

try:
    import gdal, ogr, osr
except:
    from osgeo import gdal, ogr, osr
gdal.UseExceptions()

EXCEPTION_WARNING = (
                             'Use Nansat(filename).')

class OptionError(Exception):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            'nansat.tools.OptionError will be disabled in Nansat 1.1. Use ValueError instead.',
            NansatFutureWarning)


class ProjectionError(Exception):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            'nansat.tools.ProjectionError will be disabled in Nansat 1.1. Use ' \
                'nansat.exceptions.NansatProjectionError instead.',
            NansatFutureWarning)


class GDALError(Exception):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            'nansat.tools.GDALError will be disabled in Nansat 1.1. Use ' \
                'nansat.exceptions.NansatGDALError instead.',
            NansatFutureWarning)


class NansatReadError(Exception):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            'nansat.tools.NansatReadError will be disabled in Nansat 1.1. Use ' \
                'nansat.exceptions.NansatReadError instead.',
        NansatFutureWarning)


class GeolocationError(Exception):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            'nansat.tools.GeolocationError will be disabled in Nansat 1.1. Use ' \
                'nansat.exceptions.NansatGeolocationError instead.',
            NansatFutureWarning)


class WrongMapperError(Exception):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            'nansat.tools.WrongMapperError will be disabled in Nansat 1.1. Use ' \
                'nansat.exceptions.WrongMapperError instead.',
        NansatFutureWarning)



def remove_keys(dict, keys):
    if keys is None:
        keys = []
    for key in keys:
        dict.pop(key, None)
    return dict

def register_colormaps():
    ''' Create custom colormaps and register them '''
    obpg = {'red': [(0.00, 0.56, 0.56),
                    (0.19, 0.00, 0.00),
                    (0.38, 0.00, 0.00),
                    (0.50, 0.00, 0.00),
                    (0.63, 1.00, 1.00),
                    (0.88, 1.00, 1.00),
                    (1.00, 0.40, 0.40)],

            'green': [(0.00, 0.00, 0.00),
                      (0.19, 0.00, 0.00),
                      (0.38, 1.00, 1.00),
                      (0.50, 1.00, 1.00),
                      (0.63, 1.00, 1.00),
                      (0.88, 0.00, 0.00),
                      (1.00, 0.00, 0.00)],

            'blue': [(0.00, 0.43, 0.43),
                     (0.19, 1.00, 1.00),
                     (0.38, 1.00, 1.00),
                     (0.50, 0.00, 0.00),
                     (0.63, 0.00, 0.00),
                     (0.88, 0.00, 0.00),
                     (1.00, 0.00, 0.00)],
            }


    ak01 = {'red': [(0, 0.1, 0.1,),
                    (0.1, 0.56, 0.56,),
                    (0.22, 0, 0,),
                    (0.27, 0, 0,),
                    (0.37, 0.3, 0.3,),
                    (0.47, 0, 0,),
                    (0.52, 0, 0,),
                    (0.64, 1, 1,),
                    (0.76, 1, 1,),
                    (0.88, 0.4, 0.4,),
                    (1, 1, 1,)],

            'green': [(0, 0, 0,),
                      (0.1, 0, 0,),
                      (0.22, 0, 0,),
                      (0.27, 0, 0,),
                      (0.37, 0.6, 0.6,),
                      (0.47, 0.6, 0.6,),
                      (0.52, 1, 1,),
                      (0.64, 1, 1,),
                      (0.76, 0, 0,),
                      (0.88, 0, 0,),
                      (1, 0.5, 0.5,)],

            'blue': [(0, 0.1, 0.1,),
                     (0.1, 0.5, 0.5,),
                     (0.22, 0.5, 0.5,),
                     (0.27, 1, 1,),
                     (0.37, 1, 1,),
                     (0.47, 0, 0,),
                     (0.52, 0, 0,),
                     (0.64, 0, 0,),
                     (0.76, 0, 0,),
                     (0.88, 0, 0,),
                     (1, 0.5, 0.5,)],
            }

    if MATPLOTLIB_IS_INSTALLED:
        cm.register_cmap(name='obpg', data=obpg, lut=256)
        cm.register_cmap(name='ak01', data=ak01, lut=256)

def initial_bearing(lon1, lat1, lon2, lat2):
        '''Initial bearing when traversing from point1 (lon1, lat1)
        to point2 (lon2, lat2)

        See http://www.movable-type.co.uk/scripts/latlong.html

        Parameters
        ----------
        lon1, lat1 : float
            longitude and latitude of start point
        lon2, lat2 : float
            longitude and latitude of end point

        Returns
        -------
        initial_bearing : float
            The initial bearing (azimuth direction) when heading out
            from the start point towards the end point along a great circle.

        '''
        rlon1 = np.radians(lon1)
        rlat1 = np.radians(lat1)
        rlon2 = np.radians(lon2)
        rlat2 = np.radians(lat2)
        bearing = np.arctan2(np.sin(rlon2 - rlon1) * np.cos(rlat2),
                             np.cos(rlat1) * np.sin(rlat2) -
                             np.sin(rlat1) * np.cos(rlat2) *
                             np.cos(rlon2 - rlon1))
        return np.mod(np.degrees(bearing) + 360, 360)


def haversine(lon1, lat1, lon2, lat2):
    """
    Calculate the great circle distance between two points
    on the spherical earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])
    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = np.sin(dlat/2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2)**2
    c = 2 * np.arcsin(np.sqrt(a))
    distance_meters = 6367000 * c
    return distance_meters


def add_logger(logName='', logLevel=None):
    ''' Creates and returns logger with default formatting for Nansat

    Parameters
    -----------
    logName : string, optional
        Name of the logger

    Returns
    --------
    logging.logger

    See also
    --------
    http://docs.python.org/howto/logging.html

    '''
    if logLevel is not None:
        os.environ['LOG_LEVEL'] = str(logLevel)
    # create (or take already existing) logger
    # with default logging level WARNING
    logger = logging.getLogger(logName)
    logger.setLevel(int(os.environ['LOG_LEVEL']))

    # if logger already exits, default stream handler has been already added
    # otherwise create and add a new handler
    if len(logger.handlers) == 0:
        # create console handler and set level to debug
        ch = logging.StreamHandler()
        # create formatter
        formatter = logging.Formatter('%(asctime)s|%(levelno)s|%(module)s|'
                                      '%(funcName)s|%(message)s',
                                      datefmt='%I:%M:%S')
        # add formatter to ch
        ch.setFormatter(formatter)
        # add ch to logger
        logger.addHandler(ch)

    logger.handlers[0].setLevel(int(os.environ['LOG_LEVEL']))

    return logger


def get_random_color(c0=None, minDist=100, low=0, high=255):
    ''' Create random color which is far enough from the input color

    Parameters
    ----------
        c0 : str
            hexademical representation of the color (e.g. '#ff0000' for red)
        minDist : int
            minimal distance to input color

    Returns
    -------
        c0 : str
            hexademical representation of the new random color
    '''
    if not MATPLOTLIB_IS_INSTALLED:
        raise ImportError('Matplotlib is not installed')

    # check inputs
    if c0 is None:
        c0 = '#000000'
    # convert input color to tuple of R,G,B
    c0rgb = np.array(hex2color(c0))

    # create new random color
    c1rgb = np.array([np.random.randint(low, high),
                      np.random.randint(low, high),
                      np.random.randint(low, high)])

    # calculate distance
    d = np.sum((c0rgb - c1rgb)**2)**0.5

    # if distance is small, create new random color
    if d < minDist:
        c1 = get_random_color(c0, minDist)
    else:
        # convert to HEX code
        c1 = '#%02x%02x%02x' % tuple(c1rgb)

    return c1


def parse_time(time_string):
    ''' Parse time string accounting for possible wrong formatting
    Parameters
    ----------
    time_string : str
        string with date and time
    Returns
    -------
        time_value : datetime object

    '''
    time_string = time_string.strip()
    # To account for datasets on the format YYYY-MM-DDZ which is
    # invalid since it has no time, but a timezone
    try:
        time_value = parse(time_string)
    except ValueError:
        if (len(time_string) == 11 and
                time_string.endswith('Z')):
            time_value = parse(time_string[:10])

    return time_value


def test_openable(fname):
    try:
        f = open(fname, 'r')
    except IOError:
        raise
    f.close()


def write_domain_map(border, out_filename, lon_vec=None, lat_vec=None, lon_border=10.,
                     lat_border=10.,figure_size=(6, 6), dpi=50, projection='cyl', resolution='c',
                     continets_color='coral',meridians=10, parallels=10, p_color='r', p_line='k',
                     p_alpha=0.5, padding=0., mer_labels=[False, False, False, False],
                     par_labels=[False, False, False, False], pltshow=False, labels=None):
        """Create an image with a map of the domain

        Uses Basemap to create a World Map
        Adds a semitransparent patch with outline of the Domain
        Writes to an image file

        Parameters
        -----------
        out_filename : string
            name of the output file name
        lon_vec : [floats] or [[floats]]
            longitudes of patches to display
        lat_vec : [floats] or [[floats]]
            latitudes of patches to display
        lon_border : float
            10, horisontal border around patch (degrees of longitude)
        lat_border : float
            10, vertical border around patch (degrees of latitude)
        figure_size : tuple of two integers
            (6, 6), size of the generated figure in inches
        dpi: int
            50, resolution of the output figure (size 6,6 and dpi 50
            produces 300 x 300 figure)
        projection : string, one of Basemap projections
            'cyl', projection of the map
        resolution : string, resolution of the map
            'c', crude
            'l', low
            'i', intermediate
            'h', high
            'f', full
        continets_color : string or any matplotlib color representation
            'coral', color of continets
        meridians : int
            10, number of meridians to draw
        parallels : int
            10, number of parallels to draw
        p_color : string or any matplotlib color representation
            'r', color of the Domain patch
        p_line : string or any matplotlib color representation
            'k', color of the Domain outline
        p_alpha : float 0 - 1
            0.5, transparency of Domain patch
        padding : float
            0., width of white padding around the map
        mer_labels : list of 4 booleans
            where to put meridian labels, see also Basemap.drawmeridians()
        par_lables : list of 4 booleans
            where to put parallel labels, see also Basemap.drawparallels()
        labels : list of str
            labels to print on top of patches
        """
        if not BASEMAP_LIB_IS_INSTALLED:
            raise ImportError(' Basemap is not installed. Cannot use Domain.write_map. '
                              ' Enable by: conda install -c conda forge basemap ')

        # if lat/lon vectors are not given as input
        if lon_vec is None or lat_vec is None or len(lon_vec) != len(lat_vec):
            lon_vec, lat_vec = border

        # convert vectors to numpy arrays
        lon_vec = np.array(lon_vec)
        lat_vec = np.array(lat_vec)

        # estimate mean/min/max values of lat/lon of the shown area
        # (real lat min max +/- latBorder) and (real lon min max +/- lonBorder)
        min_lon = max(-180, lon_vec.min() - lon_border)
        max_lon = min(180, lon_vec.max() + lon_border)
        min_lat = max(-90, lat_vec.min() - lat_border)
        max_lat = min(90, lat_vec.max() + lat_border)
        mean_lon = lon_vec.mean()
        mean_lat = lat_vec.mean()

        # generate template map (can be also tmerc)
        plt.figure(num=1, figsize=figure_size, dpi=dpi)
        bmap = Basemap(projection=projection,
                       lat_0=mean_lat, lon_0=mean_lon,
                       llcrnrlon=min_lon, llcrnrlat=min_lat,
                       urcrnrlon=max_lon, urcrnrlat=max_lat,
                       resolution=resolution)

        # add content: coastline, continents, meridians, parallels
        bmap.drawcoastlines()
        bmap.fillcontinents(color=continets_color)
        bmap.drawmeridians(np.linspace(min_lon, max_lon, meridians),
                           labels=mer_labels, fmt='%2.1f')
        bmap.drawparallels(np.linspace(min_lat, max_lat, parallels),
                           labels=par_labels, fmt='%2.1f')

        # convert input lat/lon vectors to arrays of vectors with one row
        # if only one vector was given
        if len(lon_vec.shape) == 1:
            lon_vec = [lon_vec]
            lat_vec = [lat_vec]

        for i in range(len(lon_vec)):
            # convert lat/lons to map units
            map_x, map_y = bmap(list(lon_vec[i].flat), list(lat_vec[i].flat))

            # from x/y vectors create a Patch to be added to map
            boundary = Polygon(list(zip(map_x, map_y)),
                               alpha=p_alpha, ec=p_line, fc=p_color)

            # add patch to the map
            plt.gca().add_patch(boundary)
            plt.gca().set_aspect('auto')

            if labels is not None and labels[i] is not None:
                plt.text(np.mean(map_x), np.mean(map_y), labels[i],
                         va='center', ha='right', alpha=0.5, fontsize=10)

        # save figure and close
        plt.savefig(out_filename, bbox_inches='tight',
                    dpi=dpi, pad_inches=padding)
        if pltshow:
            plt.show()
        else:
            plt.close('all')

register_colormaps()


numpy_to_gdal_type = {
    'uint8': 'Byte',
    'int8': 'Byte',
    'uint16': 'UInt16',
    'int16': 'Int16',
    'uint32': 'UInt32',
    'int32': 'Int32',
    'float32': 'Float32',
    'float64': 'Float64',
    'complex64': 'CFloat32',
    'complex128': 'CFloat64'}

gdal_type_to_offset = {
    'Byte': '1',
    'UInt16': '2',
    'Int16': '2',
    'UInt32': '4',
    'Int32': '4',
    'Float32': '4',
    'Float64': '8',
    'CFloat32': '8',
    'CFloat64': '16'}



# Name:         mapper_occci_online.py
# Purpose:      Nansat mapping for OC CCI data, stored online in THREDDS
# Author:       Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import json

import numpy as np

import pythesint as pti

from nansat.nsr import NSR
from nansat.mappers.opendap import Opendap

#https://rsg.pml.ac.uk/thredds/dodsC/CCI_ALL-v2.0-8DAY
#https://rsg.pml.ac.uk/thredds/dodsC/CCI_ALL-v2.0-MONTHLY
class Mapper(Opendap):
    ''' VRT with mapping of WKV for NCEP GFS '''
    baseURLs = ['https://rsg.pml.ac.uk/thredds/dodsC/CCI_ALL',
                'https://www.oceancolour.org/thredds/dodsC/CCI_ALL',
                'https://esgf-data1.ceda.ac.uk/thredds/dodsC/esg_esacci/ocean_colour/data/v2-release/geographic/netcdf/'
                ]
    timeVarName = 'time'
    xName = 'lon'
    yName = 'lat'
    timeCalendarStart = '1970-01-01'
    srcDSProjection = NSR().wkt

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 date=None, ds=None, bands=None, cachedir=None,
                 **kwargs):
        ''' Create NCEP VRT
        Parameters:
            filename : URL
            date : str
                2010-05-01
            ds : netCDF.Dataset
                previously opened dataset

        '''
        self.test_mapper(filename)
        self.create_vrt(filename, gdalDataset, gdalMetadata, date, ds, bands, cachedir)

        # add instrument and platform
        mm = pti.get_gcmd_instrument('Passive Remote Sensing')
        ee = pti.get_gcmd_platform('Earth Observation Satellites')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))

    def convert_dstime_datetimes(self, dsTime):
        ''' Convert time variable to np.datetime64 '''
        dsDatetimes = np.array([np.datetime64(self.timeCalendarStart) + day
                                for day in dsTime]).astype('M8[s]')

        return dsDatetimes



# -------------------------------------------------------------------------------
# Name:         mapper_topography.py
# Purpose:      Mapping for the global 30 arc-second elevation
#
# Author:       Morten Wergeland Hansen
# Modified:	Morten Wergeland Hansen
#
# Created:	04.06.2015
# Last modified:08.06.2015 10:27
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
# -------------------------------------------------------------------------------
from __future__ import division, unicode_literals, absolute_import, print_function
import re
import os.path

from nansat.vrt import VRT

from nansat.exceptions import WrongMapperError


class Mapper(VRT):

    """
    Mapping for the GTOPO30 (`<https://lta.cr.usgs.gov/GTOPO30>`_) and the GMTED2010
    (`<https://lta.cr.usgs.gov/GMTED2010>`_) global elevation models.

    Parameters:
    -----------
    filename : string
        The vrt filename, e.g., ``gtopo30.vrt``
    gdal_dataset : osgeo.gdal.Dataset
        The GDAL dataset returned by ``gdal.Open(filename)``

    Either the name of a GTOPO30 DEM file or GMTED2010 tif file, or ``<path>/<dem>.vrt``. The latter
    is an aggregation of the DEM-files available from the given DEM. The GTOPO30 vrt does not
    contain the Antarctic, because this is in polarstereographic projection.

    You can create your own ``gtopo30.vrt`` file with gdal, e.g.:

    .. code-block:: bash

        gdalbuildvrt <dem>.vrt [E,W]*.DEM

    Remember to update this mapper by adding allowed filenames to the list of accepted filenames
    (accepted_names) if you create or apply new DEM datasets.

    """

    def __init__(self, filename, gdal_dataset, *args, **kwargs):

        bn = os.path.basename(filename)
        accepted_names = [
                'gmted2010_30.vrt',
                'gtopo30.vrt',
                '.*.DEM',
                '.*_gmted_mea.*.tif',
            ]

        correct_mapper = False
        for accepted_name in accepted_names:
            m = re.search(accepted_name, filename)
            if m and m.group(0):
                correct_mapper = True
                break

        #if not bn=='gtopo30.vrt' and not os.path.splitext(bn)[1]=='.DEM':
        if not correct_mapper:
            raise WrongMapperError

        metaDict = [{'src': {'SourceFilename': filename, 'SourceBand':  1},
                     'dst': {'wkv': 'height_above_reference_ellipsoid'}}]

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdal_dataset)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)



# Name:         mapper_occci_online.py
# Purpose:      Nansat mapping for OC CCI data, stored online in THREDDS
# Author:       Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import os
import json
import datetime as dt

import numpy as np

import pythesint as pti

from nansat.nsr import NSR
from nansat.mappers.opendap import Dataset, Opendap

#http://thredds.met.no/thredds/dodsC/myocean/siw-tac/sst-metno-arc-sst03/20121001000000-METNO-L4_GHRSST-SSTfnd-METNO_OI-ARC-v02.0-fv01.0.nc
#http://thredds.met.no/thredds/dodsC/myocean/siw-tac/sst-metno-arc-sst03_V1/20120808-METNO-L4UHfnd-ARC-v01-fv01-METNO_OI.nc
#http://thredds.met.no/thredds/dodsC/sea_ice/SST-METNO-ARC-SST_L4-OBS-V2-V1/sst_arctic_aggregated


class Mapper(Opendap):
    """VRT with mapping of WKV for NCEP GFS"""
    baseURLs = ['http://thredds.met.no/thredds/dodsC/myocean/siw-tac/sst-metno-arc-sst03/',
                'http://thredds.met.no/thredds/dodsC/myocean/siw-tac/sst-metno-arc-sst03_V1/',
                'http://thredds.met.no/thredds/dodsC/sea_ice/SST-METNO-ARC-SST_L4-OBS-V2-V1/']
    timeVarName = 'time'
    xName = 'lon'
    yName = 'lat'
    t0 = dt.datetime(1981, 1, 1)
    srcDSProjection = NSR().wkt

    def __init__(self, filename, gdalDataset, gdalMetadata, date=None, ds=None, bands=None,
                 cachedir=None, **kwargs):
        ''' Create NCEP VRT
        Parameters:
            filename : URL
            date : str
                2010-05-01
            ds : netCDF.Dataset
                previously opened dataset

        '''
        self.test_mapper(filename)
        self.create_vrt(filename, gdalDataset, gdalMetadata, date, ds, bands, cachedir)

        # add instrument and platform
        mm = pti.get_gcmd_instrument('Passive Remote Sensing')
        ee = pti.get_gcmd_platform('Earth Observation Satellites')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))

    def convert_dstime_datetimes(self, dsTime):
        ''' Convert time variable to np.datetime64 '''

        dsDatetimes = np.array([np.datetime64(self.t0 + dt.timedelta(seconds=int(day)))
                                for day in dsTime]).astype('M8[s]')
        return dsDatetimes



# Name:    domain.py
# Purpose: Container of Domain class
# Authors:      Asuka Yamakawa, Anton Korosov, Knut-Frode Dagestad,
#               Morten W. Hansen, Alexander Myasoyedov,
#               Dmitry Petrenko, Evgeny Morozov, Aleksander Vines
# Created:      29.06.2011
# Copyright:    (c) NERSC 2011 - 2015
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import division, absolute_import

import re
import numpy as np
from xml.etree.ElementTree import ElementTree

from nansat.tools import add_logger, initial_bearing, haversine, gdal, osr, ogr
from nansat.tools import write_domain_map
from nansat.nsr import NSR
from nansat.vrt import VRT

import warnings
from nansat.exceptions import NansatProjectionError
from nansat.warnings import NansatFutureWarning


class Domain(object):
    """Container for geographical reference of a raster

    A Domain object describes all attributes of geographical
    reference of a raster:
      * width and height (number of pixels)
      * pixel size (e.g. in decimal degrees or in meters)
      * relation between pixel/line coordinates and geographical
        coordinates (e.g. a linear relation)
      * type of data projection (e.g. geographical or stereographic)

    Parameters
    ----------
    srs : PROJ4 or EPSG or WKT or NSR or osr.SpatialReference()
        Input parameter for nansat.NSR()
    ext : string
        some gdalwarp options + additional options
        [http://www.gdal.org/gdalwarp.html]
        Specifies extent, resolution / size
        Available options: (('-te' or '-lle') and ('-tr' or '-ts'))
        (e.g. '-lle -10 30 55 60 -ts 1000 1000' or
        '-te 100 2000 300 10000 -tr 300 200')
        -tr resolutionx resolutiony
        -ts sizex sizey
        -te xmin ymin xmax ymax
        -lle lonmin latmin lonmax latmax
    ds : GDAL dataset
    lat : Numpy array
        Grid with latitudes
    lon : Numpy array
        Grid with longitudes
    name : string, optional
        Name to be added to the Domain object
    logLevel : int, optional
        level of logging

    Examples
    --------
        >>> d = Domain(srs, ext) #size, extent and spatial reference is given by strings
        >>> d = Domain(ds=GDALDataset) #size, extent copied from input GDAL dataset
        >>> d = Domain(srs, ds=GDALDataset) # spatial reference is given by srs,
            but size and extent is determined from input GDAL dataset
        >>> d = Domain(lon=lonGrid, lat=latGrid) # Size, extent and spatial reference is given
            by two grids of longitude and latitude

    Notes
    -----
    The core of Domain is a GDAL Dataset. It has no bands, but only
    georeference information: rasterXsize, rasterYsize, GeoTransform and
    Projection or GCPs, etc. which fully describe dimentions and spatial
    reference of the grid.

    There are three ways to store geo-reference in a GDAL dataset:
      * Using GeoTransfrom to define linear relationship between raster
        pixel/line and geographical X/Y coordinates
      * Using GCPs (set of Ground Control Points) to define non-linear
        relationship between pixel/line and X/Y
      * Using Geolocation Array - full grids of X/Y coordinates for
        each pixel of a raster
    The relation between X/Y coordinates of the raster and latitude/longitude
    coordinates is defined by projection type and projection parameters.
    These pieces of information are therefore stored in Domain:
      * Type and parameters of projection +
        * GeoTransform, or
        * GCPs, or
        * GeolocationArrays

    Domain has methods for basic operations with georeference information:
      * creating georeference from input options;
      * fetching corner, border or full grids of X/Y coordinates;
      * making map of the georeferenced grid in a PNG or KML file;
      * and some more...

    The main attribute of Domain is a VRT object self.vrt.
    Nansat inherits from Domain and adds bands to self.vrt

    Raises
    -------
    NansatProjectionError : occurs when Projection() is empty
        despite it is required for creating extentDic.
    OptionError : occures when the arguments are not proper.

    See Also
    ---------
    Nansat.reproject()
    [http://www.gdal.org/gdalwarp.html]
    [http://trac.osgeo.org/proj/]
    [http://spatialreference.org/]
    [http://www.gdal.org/ogr/osr_tutorial.html]

    """
    REPROJECT_GCPS_WARNING = ('reproject_GCPs() will be disabled in Nansat 1.1. '
                             'Use reproject_gcps().')
    WRITE_MAP_WARNING = ('Domain.write_map() will be disabled in Nansat 1.1. '
                         'Use nansat.write_domain_map().')

    OUTPUT_SEPARATOR = '-' * 40 + '\n'
    KML_BASE = '''<?xml version="1.0" encoding="UTF-8"?>
    <kml xmlns="http://www.opengis.net/kml/2.2"
    xmlns:gx="http://www.google.com/kml/ext/2.2"
    xmlns:kml="http://www.opengis.net/kml/2.2"
    xmlns:atom="http://www.w3.org/2005/Atom">
    {content}
    </kml>'''

    # instance attributes
    vrt = None
    logger = None
    name = None

    def __init__(self, srs=None, ext=None, ds=None, lon=None,
                 lat=None, name='', logLevel=None):
        """Create Domain from GDALDataset or string options or lat/lon grids"""
        # set default attributes
        self.logger = add_logger('Nansat', logLevel)
        self.name = name

        self.logger.debug('ds: %s' % str(ds))
        self.logger.debug('srs: %s' % srs)
        self.logger.debug('ext: %s' % ext)

        # If too much information is given raise error
        if ds is not None and srs is not None and ext is not None:
            raise ValueError('Ambiguous specification of both dataset, srs- and ext-strings.')

        # choose between input opitons:
        # ds
        # ds and srs
        # srs and ext
        # lon and lat

        # if only a dataset is given:
        #     copy geo-reference from the dataset
        if ds is not None and srs is None:
            self.vrt = VRT.from_gdal_dataset(ds)

        # If dataset and srs are given (but not ext):
        #   use AutoCreateWarpedVRT to determine bounds and resolution
        elif ds is not None and srs is not None:
            srs = NSR(srs)
            tmp_vrt = gdal.AutoCreateWarpedVRT(ds, None, srs.wkt)
            if tmp_vrt is None:
                raise NansatProjectionError('Could not warp the given dataset to the given SRS.')
            else:
                self.vrt = VRT.from_gdal_dataset(tmp_vrt)

        # If SpatialRef and extent string are given (but not dataset)
        elif srs is not None and ext is not None:
            srs = NSR(srs)
            # create full dictionary of parameters
            extent_dict = Domain._create_extent_dict(ext)

            # convert -lle to -te
            if 'lle' in extent_dict.keys():
                extent_dict = self._convert_extentDic(srs, extent_dict)

            # get size/extent from the created extent dictionary
            geo_transform, raster_x_size, raster_y_size = self._get_geotransform(extent_dict)
            # create VRT object with given geo-reference parameters
            self.vrt = VRT.from_dataset_params(x_size=raster_x_size, y_size=raster_y_size,
                                               geo_transform=geo_transform,
                                               projection=srs.wkt,
                                               gcps=[], gcp_projection='')
        elif lat is not None and lon is not None:
            # create self.vrt from given lat/lon
            self.vrt = VRT.from_lonlat(lon, lat)
        else:
            raise ValueError('"dataset" or "srsString and extentString" '
                              'or "dataset and srsString" are required')

        self.logger.debug('vrt.dataset: %s' % str(self.vrt.dataset))

    def __repr__(self):
        """Creates string with basic info about the Domain object

        Returns
        ---------
         out_str : str
            size, projection and corner coordinates

        """
        corners_temp = '\t (%6.2f, %6.2f)  (%6.2f, %6.2f)\n'

        out_str = 'Domain:[%d x %d]\n' % self.shape()[::-1]
        out_str += self.OUTPUT_SEPARATOR
        corners = self.get_corners()
        out_str += 'Projection:\n'
        out_str += (NSR(self.vrt.get_projection()).ExportToPrettyWkt(1) + '\n')
        out_str += self.OUTPUT_SEPARATOR
        out_str += 'Corners (lon, lat):\n'
        out_str += corners_temp % (corners[0][0], corners[1][0], corners[0][2], corners[1][2])
        out_str += corners_temp % (corners[0][1], corners[1][1], corners[0][3], corners[1][3])
        return out_str

    def write_kml(self, xmlFileName=None, kmlFileName=None):
        """Write KML file with domains

        Convert XML-file with domains into KML-file for GoogleEarth
        or write KML-file with the current Domain

        Parameters
        -----------
        xmlFileName : string, optional
            Name of the XML-file to convert. If only this value is given
            - kmlFileName=xmlFileName+'.kml'

        kmlFileName : string, optional
            Name of the KML-file to generate from the current Domain

        """
        xml_filename = xmlFileName
        kml_filename = kmlFileName

        template = '''<Document>
        \t<name>{filename}</name>
        \t\t<Folder><name>{filename}</name><open>1</open>
        {borders}
        \t\t</Folder></Document>'''

        # test input options
        if xml_filename and not kml_filename:
            # if only input XML-file is given - convert it to KML

            # open XML, get all domains
            with open(xml_filename, 'rb') as xml_file:
                xml_domains = list(ElementTree(file=xml_file).getroot())

            # convert domains in XML into list of domains
            domains = [Domain(srs=xml_filename, ext=domain.attrib['name'])
                       for domain in xml_domains]

        elif not xml_filename and kml_filename:
            # if only output KML-file is given
            # then convert the current domain to KML
            domains = [self]

        else:
            # otherwise it is potentially error
            raise ValueError('Either xmlFileName(%s)\
             or kmlFileName(%s) are wrong' % (xml_filename, kml_filename))

        # get border of each domain and join them to a one string
        borders = ''.join([domain._get_border_kml() for domain in domains])
        # open KML, write the modified template
        with open(kml_filename, 'wt') as kml_file:
            kml_content = template.format(name=self.name, filename=kml_filename, borders=borders)
            kml_file.write(self.KML_BASE.format(content=kml_content))

    def _get_border_kml(self):
        """Generate Placemark entry for KML

        Returns
        --------
        kmlEntry : str
            String with the Placemark entry

        """
        klm_entry = '''\t\t\t<Placemark>
        \t\t\t\t<name>{name}</name>
        \t\t\t\t<Style>
        \t\t\t\t\t<LineStyle><color>ffffffff</color></LineStyle>
        \t\t\t\t\t<PolyStyle><fill>0</fill>'
        \t\t\t\t</Style>
        \t\t\t\t<Polygon><tessellate>1</tessellate><outerBoundaryIs><LinearRing><coordinates>
        {coordinates}
        </coordinates></LinearRing></outerBoundaryIs></Polygon></Placemark>'''
        domain_lon, domain_lat = self.get_border()
        # convert Border coordinates into KML-like string
        coordinates = ''.join(['%f,%f,0 ' % (lon, lat) for lon, lat in zip(domain_lon, domain_lat)])
        return klm_entry.format(name=self.name, coordinates=coordinates)

    def write_kml_image(self, kmlFileName, kmlFigureName=None):
        """Create KML file for already projected image

        Write Domain Image into KML-file for GoogleEarth

        Parameters
        -----------
        kmlFileName : str
            Name of the KML-file to generate from the current Domain
        kmlFigureName : str
            Name of the projected image stored in .png format

        Examples
        ---------
            >>> n.undo(100) # cancel previous reprojection
            >>> lons, lats = n.get_corners() # Get corners of the image and the pixel resolution
            >>> srsString = '+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs'
            >>> extentString = '-lle %f %f %f %f -ts 3000 3000'
                % (min(lons), min(lats), max(lons), max(lats))
            >>> d = Domain(srs=srsString, ext=extentString) # Create Domain with
                stereographic projection, corner coordinates and resolution 1000m
            >>> n.reproject(d)
            >>> n.write_figure(filename=figureName, bands=[3], clim=[0,0.15],
                                cmapName='gray', transparency=0)
            >>> n.write_kml_image(kmlFileName=oPath + filename + '.kml',
                                kmlFigureName=figureName) # 6.

        """

        kml_filename = kmlFileName
        kml_figurename = kmlFigureName

        template = '''<GroundOverlay>
        \t<name>{filename}</name>
        \t<Icon>
        \t\t<href>{figurename}</href>
        \t\t<viewBoundScale>0.75</viewBoundScale>
        \t</Icon>
        \t<LatLonBox>
        \t\t<north>{north}</north>
        \t\t<south>{south}</south>
        \t\t<east>{east}</east>
        \t\t<west>{west}</west>
        \t</LatLonBox>
        </GroundOverlay>'''
        # test input options
        if kml_figurename is None:
            raise ValueError('kmlFigureName(%s) is not specified' % kmlFigureName)

        # get corner of the domain and add to KML
        domain_lon, domain_lat = self.get_corners()
        with open(kml_filename, 'wt') as kml_file:
            kml_content = template.format(filename=kml_filename, figurename=kml_figurename,
                                          north=max(domain_lat), south=min(domain_lat),
                                          east=max(domain_lon), west=min(domain_lon))
            kml_file.write(self.KML_BASE.format(content=kml_content))

    def get_geolocation_grids(self, stepSize=1, dstSRS=NSR()):
        """Get longitude and latitude grids representing the full data grid

        If GEOLOCATION is not present in the self.vrt.dataset then grids
        are generated by converting pixel/line of each pixel into lat/lon
        If GEOLOCATION is present in the self.vrt.dataset then grids are read
        from the geolocation bands.

        Parameters
        -----------
        stepSize : int
            Reduction factor if output is desired on a reduced grid size

        Returns
        --------
        longitude : numpy array
            grid with longitudes
        latitude : numpy array
            grid with latitudes
        """
        step_size = stepSize
        dst_srs = dstSRS
        x_vec = list(range(0, self.vrt.dataset.RasterXSize, step_size))
        y_vec = list(range(0, self.vrt.dataset.RasterYSize, step_size))
        x_grid, y_grid = np.meshgrid(x_vec, y_vec)

        if hasattr(self.vrt, 'geolocation') and len(self.vrt.geolocation.data) > 0:
            # if the vrt dataset has geolocationArray
            # read lon,lat grids from geolocationArray
            lon_grid, lat_grid = self.vrt.geolocation.get_geolocation_grids()
            lon_arr, lat_arr = lon_grid[y_grid, x_grid], lat_grid[y_grid, x_grid]
        else:
            # generate lon,lat grids using GDAL Transformer
            lon_vec, lat_vec = self.transform_points(x_grid.flatten(), y_grid.flatten(),
                                                     dstSRS=dst_srs)
            lon_arr = lon_vec.reshape(x_grid.shape)
            lat_arr = lat_vec.reshape(x_grid.shape)

        return lon_arr, lat_arr

    def _convert_extentDic(self, dstSRS, extentDic):
        """Convert -lle option (lat/lon) to -te (proper coordinate system)

        Source SRS from LAT/LON projection and target SRS from dstWKT.
        Create osr.CoordinateTransformation based on these SRSs and
        convert given values in degrees to the destination coordinate
        system given by WKT.
        Add key 'te' and the converted values into the extentDic.

        Parameters
        -----------
        dstSRS : NSR
            Destination Spatial Reference
        extentDic : dictionary
            dictionary with 'lle' key

        Returns
        --------
        extentDic : dictionary
            input dictionary + 'te' key and its values

        """
        coorTrans = osr.CoordinateTransformation(NSR(), dstSRS)

        # convert lat/lon given by 'lle' to the target coordinate system and
        # add key 'te' and the converted values to extentDic
        x1, y1, _ = coorTrans.TransformPoint(extentDic['lle'][0], extentDic['lle'][3])
        x2, y2, _ = coorTrans.TransformPoint(extentDic['lle'][2], extentDic['lle'][3])
        x3, y3, _ = coorTrans.TransformPoint(extentDic['lle'][2], extentDic['lle'][1])
        x4, y4, _ = coorTrans.TransformPoint(extentDic['lle'][0], extentDic['lle'][1])

        minX = min([x1, x2, x3, x4])
        maxX = max([x1, x2, x3, x4])
        minY = min([y1, y2, y3, y4])
        maxY = max([y1, y2, y3, y4])

        extentDic['te'] = [minX, minY, maxX, maxY]

        return extentDic

    @staticmethod
    def _add_to_dict(extent, option):
        """Convert options to list of float values and add to <extent> dict"""
        try:
            parameters = [float(el.strip()) for el in option[1:]]
        except ValueError:
            raise ValueError('Input values must be int or float')

        key = option[0].strip().replace('-', '')
        extent[key] = parameters
        return key, extent

    @staticmethod
    def _validate_ts_tr(options):
        example = '<-tr x_resolution y_resolution> or <-ts width height>'
        Domain._check_size(len(options), 2, ('-ts', '-tr'), example)
        if options[0] <= 0 or options[1] <= 0:
            raise ValueError('Resolution or width and height must be bigger than 0: %s' % example)

    @staticmethod
    def _validate_te_lle(options):
        example = '<-te x_min y_min x_max y_max> or <-lle min_lon min_lat max_lon max_lat>'
        Domain._check_size(len(options), 4, ('-te', '-lle'), example)
        if options[0] >= options[2] or options[1] >= options[3]:
            raise ValueError('Min cannot be bigger than max: %s' % example)

    @staticmethod
    def _check_size(params_len, size, names, example):
        if params_len != size:
            raise ValueError('%s and %s requires exactly %s parameters (%s given): %s'
                              % (names[0], names[1], size, params_len, example))

    @staticmethod
    def _gen_regexp(param_1, param_2, size):
        return '(-%s|-%s)%s\s?' % (param_1, param_2, '(\s+[-+]?\d*[.\d*]*)' * size)

    @staticmethod
    def _create_extent_dict(extent_str):
        """Create a dictionary from extentString

        Check if extentString is proper.
            * '-te' and '-lle' take 4 numbers.
            * '-ts' and '-tr' take 2 numbers.
            * the combination should be ('-te' or '-lle') and ('-ts' or '-tr')
        If it is proper, create a dictionary
        Otherwise, raise the error.

        Parameters
        -----------
        extentString : string
            '-te xMin yMin xMax yMax',
            '-tr xResolution yResolution',
            '-ts width height',
            '-lle minlon minlat maxlon maxlat'

        Returns
        --------
        extentDic : dictionary
            has key ('te' or 'lle') and ('tr' or 'ts') and their values.

        Raises
        -------
        ValueError : occurs when the extent_str is improper

        """

        combinations = [('te', 'lle', 4), ('ts', 'tr', 2)]
        extent_dict = {}
        for combination in combinations:
            try:
                option = re.findall(Domain._gen_regexp(*combination), extent_str)[0]
            except IndexError:
                raise ValueError('<extent_dict> must contains exactly 2 parameters '
                                  '("-te" or "-lle") and ("-ts" or "-tr")')
            key, extent_dict = Domain._add_to_dict(extent_dict, option)
            if key is 'te' or key is 'lle':
                Domain._validate_te_lle(extent_dict[key])
            elif key is 'ts' or key is 'tr':
                Domain._validate_ts_tr(extent_dict[key])

        return extent_dict

    def get_border(self, nPoints=10):
        """Generate two vectors with values of lat/lon for the border of domain

        Parameters
        -----------
        nPoints : int, optional
            Number of points on each border

        Returns
        --------
        lonVec, latVec : lists
            vectors with lon/lat values for each point at the border

        """
        n_points = nPoints
        x_size, y_size = self.shape()[::-1]
        x_rc_vec = Domain._get_row_col_vector(x_size, n_points)
        y_rc_vec = Domain._get_row_col_vector(y_size, n_points)
        col_vec, row_vec = Domain._compound_row_col_vectors(x_size, y_size, x_rc_vec, y_rc_vec)
        return self.transform_points(col_vec, row_vec)

    @staticmethod
    def _compound_row_col_vectors(x_size, y_size, x_vec, y_vec):
        col_vec = (x_vec + [x_size] * len(y_vec) + x_vec[::-1] + [0] * len(y_vec))
        row_vec = ([0] * len(x_vec) + y_vec + [y_size] * len(x_vec) + y_vec[::-1])
        return col_vec, row_vec

    @staticmethod
    def _get_row_col_vector(raster_size, n_points):
        # Int because py3 division returns float
        step = max(1, int(raster_size / n_points))
        rc_vec = list(range(0, raster_size, step))[0:n_points]
        rc_vec.append(raster_size)
        return rc_vec

    def get_border_wkt(self, *args, **kwargs):
        """Creates string with WKT representation of the border polygon

        Returns
        --------
        WKTPolygon : string
            string with WKT representation of the border polygon

        """
        lon_vec, lat_vec = self.get_border(*args, **kwargs)

        ''' The following causes erratic geometry when using
        WKTReader().read(n.get_border_wkt(nPoints=1000)) - only commented out
        now since this may cause other problems...
        '''
        warnings.warn("> 180 deg correction to longitudes - disabled..")
        polygon_border = ','.join('%s %s' % (lon, lat) for lon, lat in zip(lon_vec, lat_vec))
        # outer quotes have to be double and inner - single!
        # wktPolygon = "PolygonFromText('POLYGON((%s))')" % polyCont
        wkt = 'POLYGON((%s))' % polygon_border
        return wkt

    def get_border_geometry(self, *args, **kwargs):
        """ Get OGR Geometry of the border Polygon

        Returns
        -------
        OGR Geometry, type Polygon

        """

        return ogr.CreateGeometryFromWkt(self.get_border_wkt(*args, **kwargs))

    def overlaps(self, anotherDomain):
        """ Checks if this Domain overlaps another Domain

        Returns
        -------
        overlaps : bool
            True if Domains overlaps, False otherwise

        """

        return self.get_border_geometry().Overlaps(anotherDomain.get_border_geometry())

    def intersects(self, anotherDomain):
        """ Checks if this Domain intersects another Domain

        Returns
        -------
        intersects : bool
            True if Domains intersects, False otherwise

        """

        return self.get_border_geometry().Intersects(anotherDomain.get_border_geometry())

    def contains(self, anotherDomain):
        """ Checks if this Domain fully covers another Domain

        Returns
        -------
        contains : bool
            True if this Domain fully covers another Domain, False otherwise

        """

        return self.get_border_geometry().Contains(anotherDomain.get_border_geometry())

    def get_border_postgis(self):
        """ Get PostGIS formatted string of the border Polygon

        Returns
        -------
        str : 'PolygonFromText(PolygonWKT)'

        """

        return "PolygonFromText('%s')" % self.get_border_wkt()

    def get_corners(self):
        """Get coordinates of corners of the Domain

        Returns
        --------
        lonVec, latVec : lists
            vectors with lon/lat values for each corner

        """

        col_vec = [0, 0, self.vrt.dataset.RasterXSize, self.vrt.dataset.RasterXSize]
        row_vec = [0, self.vrt.dataset.RasterYSize, 0, self.vrt.dataset.RasterYSize]
        return self.transform_points(col_vec, row_vec)

    def get_min_max_lon_lat(self):
        """Get minimum and maximum of longitude and latitude geolocation grids

        Returns
        --------
        min_lon, max_lon, min_lat, max_lat, : float
            min/max lon/lat values for the Domain

        """
        lon_grd, lat_grd = self.get_geolocation_grids()
        return lon_grd.min(), lon_grd.max(), lat_grd.min(), lat_grd.max(),

    def get_pixelsize_meters(self):
        """Returns the pixelsize (deltaX, deltaY) of the domain

        For projected domains, the exact result which is constant
        over the domain is returned.
        For geographic (lon-lat) projections, or domains with no geotransform,
        the haversine formula is used to calculate the pixel size
        in the center of the domain.

        Returns
        --------
        delta_x, delta_y : float
            pixel size in X and Y directions given in meters
        """

        srs = osr.SpatialReference(self.vrt.dataset.GetProjection())
        if srs.IsProjected:
            if srs.GetAttrValue('unit') == 'metre':
                geoTransform = self.vrt.dataset.GetGeoTransform()
                delta_x = abs(geoTransform[1])
                delta_y = abs(geoTransform[5])
                return delta_x, delta_y

        # Estimate pixel size in center of domain using haversine formula
        center_col = round(self.vrt.dataset.RasterXSize/2)
        center_row = round(self.vrt.dataset.RasterYSize/2)
        lon00, lat00 = self.transform_points([center_col], [center_row])
        lon01, lat01 = self.transform_points([center_col], [center_row + 1])
        lon10, lat10 = self.transform_points([center_col + 1], [center_row])

        delta_x = haversine(lon00, lat00, lon01, lat01)
        delta_y = haversine(lon00, lat00, lon10, lat10)
        return delta_x[0], delta_y[0]

    @staticmethod
    def _get_geotransform(extent_dict):
        """Get the new coordinates and raster size are calculated based on the given extentDic.

        Parameters
        -----------
        extent_dict : dictionary
            includes 'te' key and 'ts' or 'tr' key

        Raises
        -------
        ValueError : occurs when maxX - minX < 0 or maxY - minY < 0

        Returns
        --------
        coordinate : list with 6 float
            GeoTransform

        raster_x_size and raster_y_size

        """
        width = extent_dict['te'][2] - extent_dict['te'][0]
        height = extent_dict['te'][3] - extent_dict['te'][1]

        if 'tr' in extent_dict.keys():
            resolution_x, resolution_y, raster_x_size, raster_y_size = \
                Domain._transform_tr(width, height, extent_dict['tr'])
        else:
            resolution_x, resolution_y, raster_x_size, raster_y_size = \
                Domain._transform_ts(width, height, extent_dict['ts'])

        # create a list for GeoTransform
        coordinates = [extent_dict['te'][0], resolution_x, 0.0,
                       extent_dict['te'][3], 0.0, resolution_y]

        return coordinates, int(raster_x_size), int(raster_y_size)

    @staticmethod
    def _transform_tr(width, height, tr_arr):
        """Calculate X and Y resolution and raster sizes from the "-tr" parameter

        Parameters
        -----------
        width : float, width of domain calculated from the "-te" extent parameter
        height: float, height of domain calculated from the "-te" extent parameter
        tr_arr: list, [<x_resolution>, <y_resolution>]

        Raises
        -------
        ValueError : occurs when the given resolution is larger than width or height.

        Returns
        --------
        resolution_x, resolution_y, raster_x_size, raster_y_size : float
        """
        resolution_x = tr_arr[0]
        resolution_y = -(tr_arr[1])

        if width < resolution_x or height < resolution_y:
            raise ValueError('"-tr" is too large. width is %s, height is %s ' % (width, height))

        raster_x_size = width / resolution_x
        raster_y_size = abs(height / resolution_y)

        return resolution_x, resolution_y, raster_x_size, raster_y_size

    @staticmethod
    def _transform_ts(width, height, ts_arr):
        raster_x_size, raster_y_size = ts_arr
        resolution_x = width / raster_x_size
        resolution_y = -abs(height / raster_y_size)

        return resolution_x, resolution_y, raster_x_size, raster_y_size

    def transform_points(self, colVector, rowVector, DstToSrc=0, dstSRS=NSR()):
        """Transform given lists of X,Y coordinates into lon/lat or inverse

        Parameters
        -----------
        colVector : lists
            X and Y coordinates in pixel/line or lon/lat  coordinate system
        DstToSrc : 0 or 1
            0 - forward transform (pix/line => lon/lat)
            1 - inverse transformation
        dstSRS : NSR
            destination spatial reference

        Returns
        --------
        X, Y : lists
            X and Y coordinates in lon/lat or pixel/line coordinate system

        """
        return self.vrt.transform_points(colVector, rowVector, dst2src=DstToSrc, dst_srs=dstSRS)

    def azimuth_y(self, reductionFactor=1):
        """Calculate the angle of each pixel position vector with respect to
        the Y-axis (azimuth).

        In general, azimuth is the angle from a reference vector (e.g., the
        direction to North) to the chosen position vector. The azimuth
        increases clockwise from direction to North.
        http://en.wikipedia.org/wiki/Azimuth

        Parameters
        -----------
        reductionFactor : integer
            factor by which the size of the output array is reduced

        Returns
        -------
        azimuth : numpy array
            Values of azimuth in degrees in range 0 - 360

        """

        lon_grd, lat_grd = self.get_geolocation_grids(reductionFactor)
        a = initial_bearing(lon_grd[1:, :], lat_grd[1:, :], lon_grd[:-1:, :], lat_grd[:-1:, :])
        # Repeat last row once to match size of lon-lat grids
        a = np.vstack((a, a[-1, :]))
        return a

    def shape(self):
        """Return Numpy-like shape of Domain object (ySize, xSize)

        Returns
        --------
        shape : tuple of two INT
            Numpy-like shape of Domain object (ySize, xSize)

        """
        return self.vrt.dataset.RasterYSize, self.vrt.dataset.RasterXSize

    def write_map(self, outputFileName, lonVec=None, latVec=None, lonBorder=10., latBorder=10.,
                  figureSize=(6, 6), dpi=50, projection='cyl', resolution='c',
                  continetsColor='coral', meridians=10, parallels=10, pColor='r', pLine='k',
                  pAlpha=0.5, padding=0., merLabels=[False, False, False, False],
                  parLabels=[False, False, False, False], pltshow=False, labels=None):
        """Create an image with a map of the domain"""
        warnings.warn(self.WRITE_MAP_WARNING, NansatFutureWarning)
        border = self.get_border()
        write_domain_map(border, outputFileName, lon_vec=lonVec, lat_vec=latVec,
                         lon_border=lonBorder, lat_border=latBorder, figure_size=figureSize,
                         dpi=dpi, projection=projection, resolution=resolution,
                         continets_color=continetsColor, meridians=meridians, parallels=parallels,
                         p_color=pColor, p_line=pLine, p_alpha=pAlpha,
                         padding=padding, mer_labels=merLabels, par_labels=parLabels,
                         pltshow=pltshow, labels=labels)

    def reproject_gcps(self, srs_string=''):
        """Reproject all GCPs to a new spatial reference system

        Necessary before warping an image if the given GCPs
        are in a coordinate system which has a singularity
        in (or near) the destination area (e.g. poles for lonlat GCPs)

        Parameters
        ----------
        srs_string : string
            SRS given as Proj4 string. If empty '+proj=stere' is used

        Notes
        --------
            Reprojects all GCPs to new SRS and updates GCPProjection
        """
        if srs_string == '':
            lon, lat = self.get_border()
            srs_string = '+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=%f +lon_0=%f +no_defs' \
                         % (np.nanmedian(lat), np.nanmedian(lon))
        self.vrt.reproject_GCPs(srs_string)

    def reproject_GCPs(self, srsString=''):
        warnings.warn(self.REPROJECT_GCPS_WARNING, NansatFutureWarning)
        self.reproject_gcps(srs_string=srsString)



#------------------------------------------------------------------------------
# Name:         test_domain.py
# Purpose:      Test the Domain class
#
# Author:       Anton Korosov
#
# Created:      29.09.2014
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
#------------------------------------------------------------------------------
import unittest
import os
import numpy as np

try:
    if 'DISPLAY' not in os.environ:
        import matplotlib; matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    from mpl_toolkits.basemap import Basemap
except ImportError:
    BASEMAP_LIB_IS_INSTALLED = False
else:
    BASEMAP_LIB_IS_INSTALLED = True

from nansat.nsr import NSR
from nansat.vrt import VRT
from nansat.domain import Domain
from nansat.tools import gdal, ogr
from nansat.figure import Image
import sys
from nansat.tests import nansat_test_data as ntd
from mock import patch, PropertyMock, Mock, MagicMock, DEFAULT

from nansat.exceptions import NansatProjectionError


EXTENT_TE_TS = "-te 25 70 35 72 -ts 500 500"
EXTENT_DICT_TE_TS = {'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]}
EXTENT_LLE_TS = "-lle 25 70 35 72 -ts 500 500"
EXTENT_DICT_LLE_TS = {'lle': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]}
EXTENT_DICT_LLE_TE_TS = {'lle': [25.0, 70.0, 35.0, 72.0],
                         'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]}
GEO_TRANSFORM = [25.0, 0.02, 0.0, 72.0, 0.0, -0.004]
RASTER_X_SIZE = 500
RASTER_Y_SIZE = 500
SRS_PROJ4 = "+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs"
SRS_EPSG = 4326
EXTENT_BERGEN = "-te 5 60 6 61 -ts 500 500"
EXTENT_WESTCOAST = "-te 1 58 6 64 -ts 500 500"
EXTENT_NORWAY = "-te 3 55 30 72 -ts 500 500"
EXTENT_PARIS = "-te 2 48 3 49 -ts 500 500"


class DomainTest(unittest.TestCase):

    def setUp(self):
        self.test_file = os.path.join(ntd.test_data_path, 'gcps.tif')
        self.test_file_projected = os.path.join(ntd.test_data_path, 'stere.tif')
        if BASEMAP_LIB_IS_INSTALLED:
            plt.switch_backend('Agg')
        if (    not os.path.exists(self.test_file)
             or not os.path.exists(self.test_file_projected) ):
            raise ValueError('No test data available')

    def test_dont_init_from_invalid_combination(self):
        self.assertRaises(ValueError, Domain)
        self.assertRaises(ValueError, Domain, None)
        with self.assertRaises(ValueError):
            Domain(ds=gdal.Open(self.test_file),
                   srs="+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs",
                   ext="-te 25 70 35 72 -ts 500 500")

    def test_init_from_GDALDataset(self):
        d = Domain(ds=gdal.Open(self.test_file))
        self.assertEqual(type(d), Domain)

    def test_init_from_GDALDataset_and_srs(self):
        d = Domain(ds=gdal.Open(self.test_file),
                   srs="+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs")
        self.assertEqual(type(d), Domain)

    @patch('nansat.domain.gdal')
    def test_dont_init_if_gdal_AutoCreateWarpedVRT_fails(self, mock_gdal):
        mock_gdal.AutoCreateWarpedVRT.return_value = None
        with self.assertRaises(NansatProjectionError):
            Domain(ds=gdal.Open(self.test_file),
                   srs="+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs")

    @patch.object(Domain, '_create_extent_dict',
        return_value={'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})
    @patch.object(Domain, '_get_geotransform',
        return_value=([25.0, 0.02, 0.0, 72.0, 0.0, -0.004], 500, 500))
    def test_init_from_srs_and_ext_te(self, mock__get_geotransform, mock__create_extent_dict):
        d = Domain(srs=4326,
                   ext="-te 25 70 35 72 -ts 500 500")
        self.assertEqual(type(d), Domain)

    @patch.object(Domain, '_create_extent_dict',
        return_value={'lle': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})
    @patch.object(Domain, '_convert_extentDic',
        return_value={'lle': [25.0, 70.0, 35.0, 72.0], 'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})
    @patch.object(Domain, '_get_geotransform',
        return_value=([25.0, 0.02, 0.0, 72.0, 0.0, -0.004], 500, 500))
    def test_init_from_srs_and_ext_lle(self, mock__get_geotransform, mock__convert_extentDic,
                                       mock__create_extent_dict):
        d = Domain(srs=4326,
                   ext="-lle 25 70 35 72 -ts 500 500")
        self.assertEqual(type(d), Domain)

    def test_init_from_lonlat(self):
        lat, lon = np.mgrid[-90:90:0.5, -180:180:0.5]
        d = Domain(lon=lon, lat=lat)
        self.assertEqual(type(d), Domain)
        self.assertEqual(d.shape(), lat.shape)

    @patch.object(Domain, 'get_corners',
        return_value=(np.array([ 25.,  25.,  35.,  35.]), np.array([ 72.,  70.,  72.,  70.])))
    def test_repr(self, mock_get_corners):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = d.__repr__()
        test = ('Domain:[500 x 500]\n'
                '----------------------------------------\n'
                'Projection:\nGEOGCS["WGS 84",\n'
                '    DATUM["WGS_1984",\n'
                '        SPHEROID["WGS 84",6378137,298.257223563]],\n'
                '    PRIMEM["Greenwich",0],\n'
                '    UNIT["degree",0.0174532925199433]]\n'
                '----------------------------------------\n'
                'Corners (lon, lat):\n'
                '\t ( 25.00,  72.00)  ( 35.00,  72.00)\n'
                '\t ( 25.00,  70.00)  ( 35.00,  70.00)\n' )
        self.assertIsInstance(result, str)
        self.assertEquals(result, test)

    def test_write_kml(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        tmpfilename = os.path.join(ntd.tmp_data_path, 'domain_write_kml.kml')
        d.write_kml(kmlFileName=tmpfilename)
        self.assertTrue(os.path.exists(tmpfilename))

    #def test__get_border_kml(self):


    #def test_write_kml_image(self):


    @patch.object(Domain, 'transform_points',
        return_value=(np.meshgrid(range(0,500),range(0,500))[0].flatten()*(35-25)/500.+25,
                      np.meshgrid(range(0,500),range(0,500))[1].flatten()*(70-72)/500.+72))
    def test_get_geolocation_grids_from_GDAL_transformer(self, mock_transform_points):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        lon, lat = d.get_geolocation_grids()
        self.assertEqual(type(lon), np.ndarray)
        self.assertEqual(type(lat), np.ndarray)
        self.assertEqual(lat.shape, (500, 500))

    def test_get_geolocation_grids_from_geolocationArray(self):
        lat, lon = np.mgrid[25:35:0.02, 70:72:0.004]
        d = Domain(lon=lon, lat=lat)
        lon, lat = d.get_geolocation_grids()
        self.assertEqual(type(lon), np.ndarray)
        self.assertEqual(type(lat), np.ndarray)
        self.assertEqual(lat.shape, (500, 500))




    def test_convert_extentDic(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = d._convert_extentDic(NSR(4326), {'lle': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})
        self.assertEqual(result, {'lle': [25.0, 70.0, 35.0, 72.0], 'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})

    def test_add_to_dict(self):
        input_1 = ['-te', '5', '60', '6', '61.1']
        output_1 = {'te': [5., 60., 6., 61.1]}
        key_1, extent_1 = Domain._add_to_dict(dict(), input_1)
        self.assertIsInstance(extent_1, dict)
        self.assertIsInstance(key_1, str)
        self.assertEqual(key_1, input_1[0].replace('-', ''))
        self.assertEqual(len(extent_1), 1)
        self.assertIsInstance(list(extent_1.values()), list)

        for el in list(extent_1.values())[0]:
            self.assertIsInstance(el, float)

        self.assertEqual(extent_1, output_1)
        input_2 = ['-te', '5', 'str', '6', '61']
        with self.assertRaises(ValueError) as opt_err:
            key_2, extent_2 = Domain._add_to_dict(dict(), input_2)
            self.assertEqual(opt_err.args[0], 'Input values must be int or float')

    def test_validate_ts_tr(self):
        input_1 = [100, 200]
        input_2 = ([0, 0], [0, 50], [10, 0], [-1, 10], [10, -100], [-100, -100])
        input_3 = [10]
        self.assertEqual(Domain._validate_ts_tr(input_1), None)
        for inp in input_2:
            with self.assertRaises(ValueError) as opt_err:
                Domain._validate_ts_tr(inp)
                self.assertEqual(opt_err.args[0], 'Resolution or width and height must be bigger '
                                                  'than 0: <-tr x_resolution y_resolution> or '
                                                  '<-ts width height>')
        with self.assertRaises(ValueError) as opt_err:
            Domain._validate_ts_tr(input_3)
            self.assertEqual(opt_err.args[0], '-ts and -tr requires exactly 2 parameters '
                                              '(1 given): <-tr x_resolution y_resolution> or '
                                              '<-ts width height>')
    def test_validate_te_lle(self):
        input_1 = [5., 60., 6., 61.]
        input_2 = ([5., 60., 5., 61.], [5., 60., 4., 61.], [5., 60., 6., 59.])
        input_3 = [60., 5., 61.]
        self.assertEqual(Domain._validate_te_lle(input_1), None)
        for inp in input_2:
            with self.assertRaises(ValueError) as opt_err:
                Domain._validate_te_lle(inp)
                self.assertEqual(opt_err.message, 'Min cannot be bigger than max: '
                                                  '<-te x_min y_min x_max y_max> or '
                                                  '<-lle min_lon min_lat max_lon max_lat>')
        with self.assertRaises(ValueError) as opt_err:
            Domain._validate_te_lle(input_3)
            self.assertEqual(opt_err.args[0], '-te and -lle requires exactly 4 parameters '
                                              '(3 given): <-te x_min y_min x_max y_max> or <-lle'
                                              ' min_lon min_lat max_lon max_lat>')

    def test_check_size(self):
        te_lle_example = '<-te x_min y_min x_max y_max> or <-lle min_lon min_lat max_lon max_lat>'
        tr_ts_example = '<-tr x_resolution y_resolution> or <-ts width height>'
        self.assertEqual(Domain._check_size(2, 2, ('-te', '-lle'), te_lle_example), None)
        with self.assertRaises(ValueError) as opt_err:
            Domain._check_size(1, 2, ('-ts', '-tr'), tr_ts_example)
            self.assertEqual(opt_err.args[0], '-ts and -tr requires exactly 2 parameters '
                                              '(1 given): <-tr x_resolution y_resolution> or '
                                              '<-ts width height>')
        with self.assertRaises(ValueError) as opt_err:
            Domain._check_size(2, 4, ('-te', '-lle'), te_lle_example)
            self.assertEqual(opt_err.args[0], '-te and -lle requires exactly 4 parameters '
                                              '(2 given): <-te x_min y_min x_max y_max> or <-lle'
                                              ' min_lon min_lat max_lon max_lat>')

    def test_gen_regexp(self):
        test_1 = '(-te|-lle)(\\s+[-+]?\\d*[.\\d*]*)(\\s+[-+]?\\d*[.\\d*]*)(\\s+[-+]?' \
                 '\\d*[.\\d*]*)(\\s+[-+]?\\d*[.\\d*]*)\\s?'
        result_1 = Domain._gen_regexp('te', 'lle', 4)
        self.assertIsInstance(result_1, str)
        self.assertEqual(result_1, test_1)
        test_2 = '(-ts|-tr)(\\s+[-+]?\\d*[.\\d*]*)(\\s+[-+]?\\d*[.\\d*]*)\\s?'
        result_2 = Domain._gen_regexp('ts', 'tr', 2)
        self.assertEqual(result_2, test_2)

    def test_create_extent_dict(self):
        test = ('-te 5 60 6 61.1 -ts 500 500',
                '-te -92.08 26.85 -92.00 26.91 -ts 200 200',
                '-te 5 60 6 61.1',
                '-te 5 60 6 61.1 -te 5 60 6 61.1')
        output_1 = {'te': [5., 60., 6., 61.1], 'ts': [500, 500]}
        output_2 = {'te': [-92.08, 26.85, -92.00, 26.91], 'ts': [200, 200]}
        result_1 = Domain._create_extent_dict(test[0])
        self.assertIsInstance(result_1, dict)
        self.assertEqual(len(list(result_1.keys())), 2)
        self.assertEqual(result_1, output_1)
        result_2 = Domain._create_extent_dict(test[1])
        self.assertEquals(result_2, output_2)
        with self.assertRaises(ValueError) as opt_err:
            Domain._create_extent_dict(test[2])
            self.assertEquals(opt_err.args[0], '<extent_dict> must contains exactly 2 parameters ')
        self.assertEqual(result_2, output_2)

        try:
            test = Domain._create_extent_dict(test[2])
        except ValueError as opt_err:
            self.assertEqual(opt_err.args[0], '<extent_dict> must contains exactly 2 parameters '
                                               '("-te" or "-lle") and ("-ts" or "-tr")')
        with self.assertRaises(ValueError) as opt_err:
            Domain._create_extent_dict(test[3])
            self.assertEquals(opt_err.args[0], '<extent_dict> must contains exactly 2 parameters '
                                               '("-te" or "-lle") and ("-ts" or "-tr")')

    def test_get_border(self):
        dom = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = dom.get_border(nPoints=10)
        lat, lon = result
        self.assertEqual(type(lat), np.ndarray)
        self.assertEqual(type(lon), np.ndarray)
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 2)
        test_x = [25., 26., 27., 28., 29., 30., 31., 32., 33., 34., 35.,
                  35., 35., 35., 35., 35., 35., 35., 35., 35., 35., 35.,
                  35., 34., 33., 32., 31., 30., 29., 28., 27., 26., 25.,
                  25., 25., 25., 25., 25., 25., 25., 25., 25., 25., 25.]
        test_y = [72., 72., 72., 72., 72., 72., 72., 72.,  72.,
                   72. ,  72. ,  72. ,  71.8,  71.6,  71.4,  71.2,  71. ,  70.8,
                   70.6,  70.4,  70.2,  70. ,  70. ,  70. ,  70. ,  70. ,  70. ,
                   70. ,  70. ,  70. ,  70. ,  70. ,  70. ,  70. ,  70.2,  70.4,
                   70.6,  70.8,  71. ,  71.2,  71.4,  71.6,  71.8,  72.]
        self.assertEqual(list(lat), test_x)
        self.assertEqual(list(lon), test_y)

    def test_compound_row_col_vectors(self):
        result = Domain._compound_row_col_vectors(30, 40, list(range(0, 33, 3)), list(range(0, 44, 4)))
        output_col, output_row = result
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 2)
        test_col = [0,  3,  6,  9,  12,  15,  18,  21,  24,  27,  30,  30,  30,  30,  30,  30,
                    30,  30,  30,  30,  30,  30,  30,  27,  24,  21,  18,  15,  12,  9,  6,  3,
                    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0]
        test_row = [0,  0,  0,  0,  0,  0,  0, 0,  0,  0,  0,  0,  4,  8,  12,  16,  20,  24,
                    28,  32,  36,  40,  40,  40,  40,  40,  40,  40,  40,  40,  40,  40,  40,
                    40,  36,  32,  28,  24,  20,  16,  12,  8,  4,  0]

        self.assertEqual(output_col, test_col)
        self.assertEqual(output_row, test_row)

    def test_get_row_col_vector(self):
        test_1 = Domain._get_row_col_vector(250, 500)
        self.assertIsInstance(test_1, list)
        self.assertEqual(test_1, list(range(251)))
        self.assertEqual(len(test_1), 251)
        test_2 = Domain._get_row_col_vector(500, 10)
        self.assertEqual(test_2, list(range(0, 550, 50)))
        self.assertEqual(len(test_2), 10 + 1)

    def test_get_border_wkt(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        bwkt = d.get_border_wkt()
        self.assertEqual(type(bwkt), str)
        self.assertTrue('POLYGON' in bwkt)

    def test_get_border_geometry(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        geom = d.get_border_geometry()
        self.assertEqual(type(geom), ogr.Geometry)

    def test_overlaps_intersects_and_contains(self):
        Bergen = Domain(4326, "-te 5 60 6 61 -ts 500 500")
        WestCoast = Domain(4326, "-te 1 58 6 64 -ts 500 500")
        Norway = Domain(4326, "-te 3 55 30 72 -ts 500 500")
        Paris = Domain(4326, "-te 2 48 3 49 -ts 500 500")

        self.assertFalse(Bergen.overlaps(Norway))
        self.assertFalse(Norway.overlaps(Bergen))
        self.assertTrue(Norway.overlaps(WestCoast))
        self.assertFalse(Paris.overlaps(Norway))

        self.assertTrue(Bergen.intersects(Norway))
        self.assertTrue(Norway.intersects(Bergen))
        self.assertTrue(Norway.intersects(WestCoast))
        self.assertFalse(Paris.intersects(Norway))

        self.assertTrue(Norway.contains(Bergen))
        self.assertFalse(Bergen.contains(Norway))
        self.assertFalse(Norway.contains(WestCoast)) # why false?
        self.assertFalse(Paris.contains(Norway))


    def test_contains(self):
        Bergen = Domain(4326, EXTENT_BERGEN)
        WestCoast = Domain(4326, EXTENT_WESTCOAST)
        Norway = Domain(4326, EXTENT_NORWAY)
        Paris = Domain(4326, EXTENT_PARIS)
        self.assertTrue(Norway.contains(Bergen))
        self.assertFalse(Bergen.contains(Norway))
        self.assertFalse(Norway.contains(WestCoast))
        self.assertFalse(Paris.contains(Norway))

    def test_transform_ts(self):
        result = Domain._transform_ts(1.5, 1.0, [750.0, 500.0])
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 4)
        for el in result:
            self.assertIsInstance(el, float)
    def test_get_border_postgis(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = d.get_border_postgis()
        self.assertIsInstance(result, str)
        self.assertEquals(result, "PolygonFromText('POLYGON((25.0 72.0,26.0 72.0,27.0 72.0,28.0 "
                                  "72.0,29.0 72.0,30.0 72.0,31.0 72.0,32.0 72.0,33.0 72.0,34.0 "
                                  "72.0,35.0 72.0,35.0 72.0,35.0 71.8,35.0 71.6,35.0 71.4,35.0 "
                                  "71.2,35.0 71.0,35.0 70.8,35.0 70.6,35.0 70.4,35.0 70.2,35.0 "
                                  "70.0,35.0 70.0,34.0 70.0,33.0 70.0,32.0 70.0,31.0 70.0,30.0 "
                                  "70.0,29.0 70.0,28.0 70.0,27.0 70.0,26.0 70.0,25.0 70.0,25.0 "
                                  "70.0,25.0 70.2,25.0 70.4,25.0 70.6,25.0 70.8,25.0 71.0,25.0 "
                                  "71.2,25.0 71.4,25.0 71.6,25.0 71.8,25.0 72.0))')")

    def test_get_corners(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        lon, lat = d.get_corners()
        self.assertTrue(all(lon - [25., 25., 35., 35.] < 0.01))
        self.assertTrue(all(lat - [72., 70., 72., 70.] < 0.01))

    def test_get_min_max_lon_lat(self):
        dom = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = dom.get_min_max_lon_lat()
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 4)
        for el in result:
            self.assertIsInstance(el, float)
        self.assertLess(result[0], result[1])
        self.assertLess(result[2], result[3])
        self.assertEqual(result, (25.0, 34.980000000000004, 70.004000000000005, 72.0))

    def test_get_pixelsize_meters(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        x, y = d.get_pixelsize_meters()
        self.assertEqual(int(x), 444)
        self.assertEqual(int(y), 723)
        d = Domain(ds=gdal.Open(self.test_file_projected))
        x, y = d.get_pixelsize_meters()
        self.assertEqual(int(x), 500)
        self.assertEqual(int(y), 500)
    def test_get_geotransform(self):
        input_1 = {'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]}
        test_1 = ([25.0, 0.02, 0.0, 72.0, 0.0, -0.004], 500, 500)
        result = Domain._get_geotransform(input_1)
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 3)
        self.assertIsInstance(result[0], list)
        self.assertEqual(len(result[0]), 6)
        for el in result[0]:
            self.assertIsInstance(el, float)
        self.assertIsInstance(result[1], int)
        self.assertIsInstance(result[2], int)
        self.assertEqual(result, test_1)

    def test_transform_tr(self):
        result = Domain._transform_tr(4.0, 1.3, [0.015, 0.005])
        self.assertIsInstance(result, tuple)
        self.assertEquals(len(result), 4)
        map(lambda el: self.assertIsInstance(el, float), result)
        with self.assertRaises(ValueError) as param_err:
            Domain._transform_tr(4.0, 1.3, [5.0, 0.005])
            self.assertEqual(param_err.message,
                             '"-tr" is too large. width is 4.0, height is 1.3 ')

    def test_transform_ts(self):
        result = Domain._transform_ts(1.5, 1.0, [750.0, 500.0])
        self.assertIsInstance(result, tuple)
        self.assertEquals(len(result), 4)
        map(lambda el: self.assertIsInstance(el, float), result)

    def test_transform_points(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        lon, lat = d.transform_points([1, 2, 3], [1, 2, 3])
        self.assertEqual(type(lon), np.ndarray)
        self.assertEqual(type(lat), np.ndarray)

    def test_transform_points_inverse(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        x, y = d.transform_points([25, 26, 27], [70, 71, 72], 1)
        self.assertTrue(all(np.round(x) == [0, 50, 100]))
        self.assertTrue(all(np.round(y) == [500, 250, 0]))

    def test_transform_points_dstsrs(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        lon, lat = d.transform_points([1, 2, 3], [1, 2, 3],
        dstSRS=NSR('+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=75 +lon_0=10 +no_defs'))
        self.assertEqual(type(lon), np.ndarray)
        self.assertEqual(type(lat), np.ndarray)

    @patch.object(Domain, 'get_geolocation_grids',
                  return_value=(np.array([[-4.        , -0.66666667,  2.66666667],
                                          [-4.        , -0.66666667,  2.66666667]]),
                                np.array([[ 7.,  7.,  7.],
                                          [ 1.,  1.,  1.]])))
    def test_azimuth_y(self, mock_get_geolocation_grids):
        d = Domain(4326, "-te -4 -5 +6 +7 -ts 3 2")
        self.assertTrue((d.azimuth_y()==np.array([[ 0.,  0.,  0.], [ 0.,  0.,  0.]])).all())

    def test_shape(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        self.assertEqual(d.shape(), (500, 500))

    @unittest.skipUnless(BASEMAP_LIB_IS_INSTALLED, 'Basemap is required')
    def test_write_map(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        tmpfilename = os.path.join(ntd.tmp_data_path, 'domain_write_map.png')
        d.write_map(tmpfilename)
        self.assertTrue(os.path.exists(tmpfilename))
        i = Image.open(tmpfilename)
        i.verify()
        self.assertEqual(i.info['dpi'], (50, 50))

    @unittest.skipUnless(BASEMAP_LIB_IS_INSTALLED, 'Basemap is required')
    def test_write_map_dpi100(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        tmpfilename = os.path.join(ntd.tmp_data_path,
                                   'domain_write_map_dpi100.png')
        d.write_map(tmpfilename, dpi=100)
        self.assertTrue(os.path.exists(tmpfilename))
        i = Image.open(tmpfilename)
        i.verify()
        self.assertEqual(i.info['dpi'], (100, 100))

    @unittest.skipUnless(BASEMAP_LIB_IS_INSTALLED, 'Basemap is required')
    def test_write_map_labels(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        tmpfilename = os.path.join(ntd.tmp_data_path,
                                   'domain_write_map_labels.png')
        d.write_map(tmpfilename,
                    merLabels=[False, False, False, True],
                    parLabels=[True, False, False, False])
        self.assertTrue(os.path.exists(tmpfilename))
        i = Image.open(tmpfilename)
        i.verify()

    def test_reproject_gcps(self):
        ds = gdal.Open(self.test_file)
        d = Domain(ds=ds)
        d.reproject_gcps('+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=75 +lon_0=10 +no_defs')
        gcp = d.vrt.dataset.GetGCPs()[0]
        self.assertTrue(gcp.GCPX > 636161)
        self.assertTrue(gcp.GCPY < -288344)

    def test_reproject_gcps_auto(self):
        ds = gdal.Open(self.test_file)
        d = Domain(ds=ds)
        d.reproject_gcps()
        gcpproj = NSR(d.vrt.dataset.GetGCPProjection())
        self.assertEqual(gcpproj.GetAttrValue('PROJECTION'),
                        'Stereographic')

    def test_reproject_GCPs(self):
        ds = gdal.Open(self.test_file)
        d = Domain(ds=ds)
        d.reproject_GCPs('+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=75 +lon_0=10 +no_defs')
        gcp = d.vrt.dataset.GetGCPs()[0]
        self.assertTrue(gcp.GCPX > 636161)
        self.assertTrue(gcp.GCPY < -288344)


    def test_repr(self):
        dom = Domain(4326, "-te 4.5 60 6 61 -ts 750 500")
        result = dom.__repr__()
        test = 'Domain:[750 x 500]\n----------------------------------------\nProjection:\nGEOGC' \
               'S["WGS 84",\n    DATUM["WGS_1984",\n        SPHEROID["WGS 84",6378137,298.257223' \
               '563]],\n    PRIMEM["Greenwich",0],\n    UNIT["degree",0.0174532925199433]]\n-----' \
               '-----------------------------------\nCorners (lon, lat):\n\t (  4.50,  61.00)  ' \
               '(  6.00,  61.00)\n\t (  4.50,  60.00)  (  6.00,  60.00)\n'

        self.assertIsInstance(result, str)
        self.assertEqual(result, test)

    @patch.multiple(Domain, get_border_geometry=DEFAULT, __init__ = Mock(return_value=None))
    def test_intersects(self, get_border_geometry):
        other_domain = MagicMock()
        d = Domain()
        d.intersects(other_domain)
        d.get_border_geometry.assert_called_once()
        d.get_border_geometry().Intersects.assert_called_once_with(other_domain.get_border_geometry())

    @patch.multiple(Domain, get_border_geometry=DEFAULT, __init__ = Mock(return_value=None))
    def test_overlaps(self, get_border_geometry):
        other_domain = MagicMock()
        d = Domain()
        d.overlaps(other_domain)
        d.get_border_geometry.assert_called_once()
        d.get_border_geometry().Overlaps.assert_called_once_with(other_domain.get_border_geometry())

if __name__ == "__main__":
    unittest.main()



# Name:         mapper_hirlam.py
# Purpose:      Nansat mapping for Hirlam model data
#               (GRIB files from www.yr.no)
# Authors:      Knut-Frode Dagestad, Morten W. Hansen
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import datetime
import json

import numpy

from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError

import pythesint as pti


class Mapper(VRT):
    ''' VRT with mapping of WKV for HIRLAM '''

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):

        try:
            geo_transform = gdalDataset.GetGeoTransform()[0:5]
        except AttributeError:
            raise WrongMapperError
        if geo_transform != (-12.1, 0.2, 0.0, 81.95, 0.0):
            raise WrongMapperError

        metaDict = [{'src': {'SourceFilename': filename,
                             'SourceBand': 2,
                             'NODATA': 9999},
                     'dst': {'wkv': 'eastward_wind',
                             'height': '10 m'}
                     },
                    {'src': {'SourceFilename': filename,
                             'SourceBand': 3,
                             'NODATA': 9999},
                     'dst': {'wkv': 'northward_wind',
                             'height': '10 m'}
                     },
                    {'src': [{'SourceFilename': filename,
                              'SourceBand': 2,
                              'DataType': gdalDataset.GetRasterBand(2).DataType
                              },
                             {'SourceFilename': filename,
                              'SourceBand': 3,
                              'DataType': gdalDataset.GetRasterBand(3).DataType
                              }],
                     'dst': {'wkv': 'wind_speed',
                             'name': 'windspeed',
                             'height': '10 m',
                             'PixelFunctionType': 'UVToMagnitude',
                             'NODATA': 9999}
                     },
                    {'src': [{'SourceFilename': filename,
                              'SourceBand': 2,
                              'DataType': gdalDataset.GetRasterBand(2).DataType
                              },
                             {'SourceFilename': filename,
                              'SourceBand': 3,
                              'DataType': gdalDataset.GetRasterBand(3).DataType
                              }],
                     'dst': {'wkv': 'wind_from_direction',
                             'name': 'winddirection',
                             'height': '10 m',
                             'PixelFunctionType': 'UVToDirectionFrom',
                             'NODATA': 9999
                             }
                     }]

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalDataset, metadata=gdalMetadata)

        # Create bands
        self.create_bands(metaDict)

        # set source, start_date, stop_date
        self.dataset.SetMetadataItem('source', 'HIRLAM')

        # Adding valid time from the GRIB file to dataset
        start_date = gdalDataset.GetRasterBand(1).GetMetadata()['GRIB_VALID_TIME']
        self.dataset.SetMetadataItem('time_coverage_start',
                datetime.datetime.utcfromtimestamp(
                int(start_date.strip().split(' ')[0])).isoformat() + '+00:00')

        stop_date = gdalDataset.GetRasterBand(gdalDataset.RasterCount).GetMetadata()['GRIB_VALID_TIME']
        self.dataset.SetMetadataItem('time_coverage_end',
                datetime.datetime.utcfromtimestamp(
                int(stop_date.strip().split(' ')[0])).isoformat() + '+00:00')

        mm = pti.get_gcmd_instrument('computer')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', 'HIgh Resolution Limited Area Model')
